import logging
import os
import subprocess
from typing import (
    Optional,
    Tuple,
)

from galaxy.tool_shed.util import basic_util
from galaxy.util import unicodify

log = logging.getLogger(__name__)

INITIAL_CHANGELOG_HASH = "000000000000"


def clone_repository(repository_clone_url: str, repository_file_dir: str, ctx_rev=None) -> Tuple[bool, Optional[str]]:
    """
    Clone the repository up to the specified changeset_revision.  No subsequent revisions will be
    present in the cloned repository.
    """
    cmd = ["hg", "clone", "--stream"]
    if ctx_rev:
        cmd.extend(["-r", str(ctx_rev)])
    cmd.extend([repository_clone_url, repository_file_dir])
    # Make sure the destination path actually exists before attempting to clone
    if not os.path.exists(repository_file_dir):
        os.makedirs(repository_file_dir)
    try:
        subprocess.check_output(cmd, stderr=subprocess.STDOUT, stdin=subprocess.DEVNULL)
        return True, None
    except Exception as e:
        error_message = f"Error cloning repository: {unicodify(e)}"
        if isinstance(e, subprocess.CalledProcessError):
            error_message += f"\nOutput was:\n{unicodify(e.output)}"
        log.error(error_message)
        return False, error_message


def copy_file_from_manifest(repo, changeset_revision, filename, dir):
    """
    Copy the latest version of the file named filename from the repository manifest to the directory
    to which dir refers.
    """
    for changeset in reversed_upper_bounded_changelog(repo, changeset_revision):
        changeset_ctx = repo[changeset]
        fctx = get_file_context_from_ctx(changeset_ctx, filename)
        if fctx and fctx not in ["DELETED"]:
            file_path = os.path.join(dir, filename)
            fh = open(file_path, "wb")
            fh.write(fctx.data())
            fh.close()
            return file_path
    return None


def get_changectx_for_changeset(repo, changeset_revision, **kwd):
    """Retrieve a specified changectx from a repository."""
    for changeset in repo.changelog:
        ctx = repo[changeset]
        if str(ctx) == changeset_revision:
            return ctx
    return None


def get_config_from_disk(config_file: str, relative_install_dir: str) -> Optional[str]:
    for root, _dirs, files in os.walk(relative_install_dir):
        if root.find(".hg") < 0:
            for name in files:
                if name == config_file:
                    return os.path.abspath(os.path.join(root, name))
    return None


def get_ctx_file_path_from_manifest(filename, repo, changeset_revision):
    """
    Get the ctx file path for the latest revision of filename from the repository manifest up
    to the value of changeset_revision.
    """
    stripped_filename = basic_util.strip_path(filename)
    for changeset in reversed_upper_bounded_changelog(repo, changeset_revision):
        manifest_ctx = repo[changeset]
        for ctx_file in manifest_ctx.files():
            ctx_file_name = basic_util.strip_path(unicodify(ctx_file))
            if ctx_file_name == stripped_filename:
                return manifest_ctx, ctx_file
    return None, None


def get_file_context_from_ctx(ctx, filename):
    """Return the mercurial file context for a specified file."""
    # We have to be careful in determining if we found the correct file because multiple files with
    # the same name may be in different directories within ctx if the files were moved within the change
    # set.  For example, in the following ctx.files() list, the former may have been moved to the latter:
    # ['tmap_wrapper_0.0.19/tool_data_table_conf.xml.sample', 'tmap_wrapper_0.3.3/tool_data_table_conf.xml.sample'].
    # Another scenario is that the file has been deleted.
    deleted = False
    filename = basic_util.strip_path(filename)
    for ctx_file in ctx.files():
        ctx_file_name = basic_util.strip_path(unicodify(ctx_file))
        if filename == ctx_file_name:
            try:
                # If the file was moved, its destination will be returned here.
                fctx = ctx[ctx_file]
                return fctx
            except LookupError:
                # Set deleted for now, and continue looking in case the file was moved instead of deleted.
                deleted = True
    if deleted:
        return "DELETED"
    return None


def pull_repository(repo_path, repository_clone_url, ctx_rev):
    """Pull changes from a remote repository to a local one."""
    try:
        subprocess.check_output(
            ["hg", "pull", "-r", ctx_rev, repository_clone_url], stderr=subprocess.STDOUT, cwd=repo_path
        )
    except Exception as e:
        error_message = f"Error pulling revision '{ctx_rev}': {unicodify(e)}"
        if isinstance(e, subprocess.CalledProcessError):
            error_message += f"\nOutput was:\n{unicodify(e.output)}"
        raise Exception(error_message)


def reversed_lower_upper_bounded_changelog(
    repo, excluded_lower_bounds_changeset_revision, included_upper_bounds_changeset_revision
):
    """
    Return a reversed list of changesets in the repository changelog after the excluded_lower_bounds_changeset_revision,
    but up to and including the included_upper_bounds_changeset_revision.  The value of excluded_lower_bounds_changeset_revision
    will be the value of INITIAL_CHANGELOG_HASH if no valid changesets exist before included_upper_bounds_changeset_revision.
    """
    # To set excluded_lower_bounds_changeset_revision, calling methods should do the following, where the value
    # of changeset_revision is a downloadable changeset_revision.
    # excluded_lower_bounds_changeset_revision = \
    #     metadata_util.get_previous_metadata_changeset_revision(app, repository, changeset_revision, downloadable=?)
    if excluded_lower_bounds_changeset_revision == INITIAL_CHANGELOG_HASH:
        appending_started = True
    else:
        appending_started = False
    reversed_changelog = []
    for changeset in repo.changelog:
        changeset_hash = str(repo[changeset])
        if appending_started:
            reversed_changelog.insert(0, changeset)
        if changeset_hash == excluded_lower_bounds_changeset_revision and not appending_started:
            appending_started = True
        if changeset_hash == included_upper_bounds_changeset_revision:
            break
    return reversed_changelog


def reversed_upper_bounded_changelog(repo, included_upper_bounds_changeset_revision):
    """
    Return a reversed list of changesets in the repository changelog up to and including the
    included_upper_bounds_changeset_revision.
    """
    return reversed_lower_upper_bounded_changelog(
        repo, INITIAL_CHANGELOG_HASH, included_upper_bounds_changeset_revision
    )


def update_repository(repo_path, ctx_rev=None):
    """
    Update the cloned repository to changeset_revision.  It is critical that the installed repository is updated to the desired
    changeset_revision before metadata is set because the process for setting metadata uses the repository files on disk.
    """
    # TODO: We may have files on disk in the repo directory that aren't being tracked, so they must be removed.
    # The codes used to show the status of files are as follows.
    # M = modified
    # A = added
    # R = removed
    # C = clean
    # ! = deleted, but still tracked
    # ? = not tracked
    # I = ignored
    # It would be nice if we could use mercurial's purge extension to remove untracked files.  The problem is that
    # purging is not supported by the mercurial API.
    cmd = ["hg", "update"]
    if ctx_rev:
        cmd.extend(["-r", ctx_rev])
    try:
        subprocess.check_output(cmd, stderr=subprocess.STDOUT, cwd=repo_path)
    except Exception as e:
        error_message = f"Error updating repository: {unicodify(e)}"
        if isinstance(e, subprocess.CalledProcessError):
            error_message += f"\nOutput was:\n{unicodify(e.output)}"
        raise Exception(error_message)


__all__ = (
    "clone_repository",
    "copy_file_from_manifest",
    "get_changectx_for_changeset",
    "get_config_from_disk",
    "get_ctx_file_path_from_manifest",
    "get_file_context_from_ctx",
    "pull_repository",
    "reversed_lower_upper_bounded_changelog",
    "reversed_upper_bounded_changelog",
    "update_repository",
)



#!/usr/bin/env python
"""This software extracts the seq, qual and ancillary information from an sff
file, like the ones used by the 454 sequencer.

Optionally, it can also split paired-end reads if given the linker sequence.
The splitting is done with maximum match, i.e., every occurence of the linker
sequence will be removed, even if occuring multiple times."""

# copyright Jose Blanca and Bastien Chevreux
# COMAV institute, Universidad Politecnica de Valencia (UPV)
# Valencia, Spain

# additions to handle paired end reads by Bastien Chevreux
# bugfixes for linker specific lengths: Lionel Guy

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

import os
import struct
import subprocess
import sys
import tempfile

__author__ = "Jose Blanca and Bastien Chevreux"
__copyright__ = "Copyright 2008, Jose Blanca, COMAV, and Bastien Chevreux"
__license__ = "GPLv3 or later"
__version__ = "0.2.10"
__email__ = "jblanca@btc.upv.es"
__status__ = "beta"

fake_sff_name = "fake_sff_name"

# readname as key: lines with matches from SSAHA, one best match
ssahapematches = {}  # type: ignore[var-annotated]
# linker readname as key: length of linker sequence
linkerlengths = {}  # type: ignore[var-annotated]

# set to true if something really fishy is going on with the sequences
stern_warning = False


def read_bin_fragment(struct_def, fileh, offset=0, data=None, byte_padding=None):
    """It reads a chunk of a binary file.

    You have to provide the struct, a file object, the offset (where to start
    reading).
    Also you can provide an optional dict that will be populated with the
    extracted data.
    If a byte_padding is given the number of bytes read will be a multiple of
    that number, adding the required pad at the end.
    It returns the number of bytes reads and the data dict.
    """
    if data is None:
        data = {}

    # we read each item
    bytes_read = 0
    for item in struct_def:
        # we go to the place and read
        fileh.seek(offset + bytes_read)
        n_bytes = struct.calcsize(item[1])
        buffer = fileh.read(n_bytes)
        read = struct.unpack(">" + item[1], buffer)
        if len(read) == 1:
            read = read[0]
        data[item[0]] = read
        bytes_read += n_bytes

    # if there is byte_padding the bytes_to_read should be a multiple of the
    # byte_padding
    if byte_padding is not None:
        pad = byte_padding
        bytes_read = ((bytes_read + pad - 1) // pad) * pad

    return (bytes_read, data)


def check_magic(magic):
    """It checks that the magic number of the file matches the sff magic."""
    if magic != 779314790:
        raise RuntimeError("This file does not seems to be an sff file.")


def check_version(version):
    """It checks that the version is supported, otherwise it raises an error."""
    if version != ("\x00", "\x00", "\x00", "\x01"):
        raise RuntimeError("SFF version not supported. Please contact the author of the software.")


def read_header(fileh):
    """It reads the header from the sff file and returns a dict with the
    information"""
    # first we read the first part of the header
    head_struct = [
        ("magic_number", "I"),
        ("version", "cccc"),
        ("index_offset", "Q"),
        ("index_length", "I"),
        ("number_of_reads", "I"),
        ("header_length", "H"),
        ("key_length", "H"),
        ("number_of_flows_per_read", "H"),
        ("flowgram_format_code", "B"),
    ]
    data = {}
    first_bytes, data = read_bin_fragment(struct_def=head_struct, fileh=fileh, offset=0, data=data)
    check_magic(data["magic_number"])
    check_version(data["version"])
    # now that we know the number_of_flows_per_read and the key_length
    # we can read the second part of the header
    struct2 = [
        ("flow_chars", str(data["number_of_flows_per_read"]) + "c"),
        ("key_sequence", str(data["key_length"]) + "c"),
    ]
    read_bin_fragment(struct_def=struct2, fileh=fileh, offset=first_bytes, data=data)
    return data


def read_sequence(header, fileh, fposition):
    """It reads one read from the sff file located at the fposition and
    returns a dict with the information."""
    # the sequence struct
    read_header_1 = [
        ("read_header_length", "H"),
        ("name_length", "H"),
        ("number_of_bases", "I"),
        ("clip_qual_left", "H"),
        ("clip_qual_right", "H"),
        ("clip_adapter_left", "H"),
        ("clip_adapter_right", "H"),
    ]

    def read_header_2(name_length):
        """It returns the struct definition for the second part of the header"""
        return [("name", str(name_length) + "c")]

    def read_data(number_of_bases):
        """It returns the struct definition for the read data section."""
        if header["flowgram_format_code"] == 1:
            flow_type = "H"
        else:
            raise Exception("file version not supported")
        number_of_bases = str(number_of_bases)
        return [
            ("flowgram_values", str(header["number_of_flows_per_read"]) + flow_type),
            ("flow_index_per_base", number_of_bases + "B"),
            ("bases", number_of_bases + "c"),
            ("quality_scores", number_of_bases + "B"),
        ]

    data = {}
    # we read the first part of the header
    bytes_read, data = read_bin_fragment(struct_def=read_header_1, fileh=fileh, offset=fposition, data=data)

    read_bin_fragment(
        struct_def=read_header_2(data["name_length"]), fileh=fileh, offset=fposition + bytes_read, data=data
    )
    # we join the letters of the name
    data["name"] = "".join(data["name"])
    offset = data["read_header_length"]
    # we read the sequence and the quality
    read_data_st = read_data(data["number_of_bases"])
    bytes_read, data = read_bin_fragment(
        struct_def=read_data_st, fileh=fileh, offset=fposition + offset, data=data, byte_padding=8
    )
    # we join the bases
    data["bases"] = "".join(data["bases"])

    # correct for the case the right clip is <= than the left clip
    # in this case, left clip is 0 are set to 0 (right clip == 0 means
    # "whole sequence")
    if data["clip_qual_right"] <= data["clip_qual_left"]:
        data["clip_qual_right"] = 0
        data["clip_qual_left"] = 0
    if data["clip_adapter_right"] <= data["clip_adapter_left"]:
        data["clip_adapter_right"] = 0
        data["clip_adapter_left"] = 0

    # the clipping section follows the NCBI's guidelines Trace Archive RFC
    # http://www.ncbi.nlm.nih.gov/Traces/trace.cgi?cmd=show&f=rfc&m=doc&s=rfc
    # if there's no adapter clip: qual -> vector
    # else:  qual-> qual
    #       adapter -> vector

    if not data["clip_adapter_left"]:
        data["clip_adapter_left"], data["clip_qual_left"] = data["clip_qual_left"], data["clip_adapter_left"]
    if not data["clip_adapter_right"]:
        data["clip_adapter_right"], data["clip_qual_right"] = data["clip_qual_right"], data["clip_adapter_right"]

    # see whether we have to override the minimum left clips
    if config["min_leftclip"] > 0:
        if data["clip_adapter_left"] > 0 and data["clip_adapter_left"] < config["min_leftclip"]:
            data["clip_adapter_left"] = config["min_leftclip"]
        if data["clip_qual_left"] > 0 and data["clip_qual_left"] < config["min_leftclip"]:
            data["clip_qual_left"] = config["min_leftclip"]

    # for handling the -c (clip) option gently, we already clip here
    #  and set all clip points to the sequence end points
    if config["clip"]:
        data["bases"], data["quality_scores"] = clip_read(data)

        data["number_of_bases"] = len(data["bases"])
        data["clip_qual_right"] = data["number_of_bases"]
        data["clip_adapter_right"] = data["number_of_bases"]
        data["clip_qual_left"] = 0
        data["clip_adapter_left"] = 0

    return data["read_header_length"] + bytes_read, data


def sequences(fileh, header):
    """It returns a generator with the data for each read."""
    # now we can read all the sequences
    fposition = header["header_length"]  # position in the file
    reads_read = 0
    while True:
        if fposition == header["index_offset"]:
            # we have to skip the index section
            fposition += header["index_length"]
            continue
        else:
            bytes_read, seq_data = read_sequence(header=header, fileh=fileh, fposition=fposition)
            yield seq_data
            fposition += bytes_read
            reads_read += 1
            if reads_read >= header["number_of_reads"]:
                break


def remove_last_xmltag_in_file(fname, tag=None):
    """Given an xml file name and a tag, it removes the last tag of the
    file if it matches the given tag. Tag removal is performed via file
    truncation.

    It the given tag is not the last in the file, a RunTimeError will be
    raised.

    The resulting xml file will be not xml valid. This function is a hack
    that allows to append records to xml files in a quick and dirty way.
    """

    fh = open(fname, "r+")
    # we have to read from the end to the start of the file and keep the
    # string enclosed by </ >
    i = -1
    last_tag = []  # the chars that form the last tag
    while True:
        fh.seek(i, 2)
        char = fh.read(1)
        if not char.isspace():
            last_tag.append(char)
        if char == "<":
            break
        i -= 1

    # we have read the last tag backwards
    last_tag = "".join(last_tag[::-1])
    # we remove the </ and >
    last_tag = last_tag.rstrip(">").lstrip("</")

    # we check that we're removing the asked tag
    if tag is not None and tag != last_tag:
        etxt = "The given xml tag (%s) was not the last one in the file" % tag
        raise RuntimeError(etxt)

    # while we are at it: also remove all white spaces in that line :-)
    i -= 1
    while True:
        fh.seek(i, 2)
        char = fh.read(1)
        if not char == " " and not char == "\t":
            break
        if fh.tell() == 1:
            break
        i -= 1

    fh.truncate()

    fh.close()
    return last_tag


def create_basic_xml_info(readname, fname):
    """Formats a number of read specific infos into XML format.
    Currently formated: name and the tags set from command line
    """
    to_print = ["    <trace>\n"]
    to_print.append("        <trace_name>")
    to_print.append(readname)
    to_print.append("</trace_name>\n")

    # extra information
    # do we have extra info for this file?
    info = None
    if config["xml_info"]:
        # with this name?
        if fname in config["xml_info"]:
            info = config["xml_info"][fname]
        else:
            # with no name?
            try:
                info = config["xml_info"][fake_sff_name]
            except KeyError:
                pass
    # we print the info that we have
    if info:
        for key in info:
            to_print.append("        <" + key + ">" + info[key] + "</" + key + ">\n")

    return "".join(to_print)


def create_clip_xml_info(readlen, adapl, adapr, quall, qualr):
    """Takes the clip values of the read and formats them into XML
    Corrects "wrong" values that might have resulted through
    simplified calculations earlier in the process of conversion
    (especially during splitting of paired-end reads)
    """

    to_print = [""]

    # if right borders are >= to read length, they don't need
    # to be printed
    if adapr >= readlen:
        adapr = 0
    if qualr >= readlen:
        qualr = 0

    # BaCh
    # when called via split_paired_end(), some values may be < 0
    #  (when clip values were 0 previously)
    # instead of putting tons of if clauses for different calculations there,
    #  I centralise corrective measure here
    # set all values <0 to 0

    if adapr < 0:
        adapr = 0
    if qualr < 0:
        qualr = 0
    if adapl < 0:
        adapl = 0
    if quall < 0:
        quall = 0

    if quall:
        to_print.append("        <clip_quality_left>")
        to_print.append(str(quall))
        to_print.append("</clip_quality_left>\n")
    if qualr:
        to_print.append("        <clip_quality_right>")
        to_print.append(str(qualr))
        to_print.append("</clip_quality_right>\n")
    if adapl:
        to_print.append("        <clip_vector_left>")
        to_print.append(str(adapl))
        to_print.append("</clip_vector_left>\n")
    if adapr:
        to_print.append("        <clip_vector_right>")
        to_print.append(str(adapr))
        to_print.append("</clip_vector_right>\n")
    return "".join(to_print)


def create_xml_for_unpaired_read(data, fname):
    """Given the data for one read it returns an str with the xml ancillary
    data."""
    to_print = [create_basic_xml_info(data["name"], fname)]
    # clippings in the XML only if we do not hard clip
    if not config["clip"]:
        to_print.append(
            create_clip_xml_info(
                data["number_of_bases"],
                data["clip_adapter_left"],
                data["clip_adapter_right"],
                data["clip_qual_left"],
                data["clip_qual_right"],
            )
        )
    to_print.append("    </trace>\n")
    return "".join(to_print)


def format_as_fasta(name, seq, qual):
    name_line = "".join((">", name, "\n"))
    seqstring = "".join((name_line, seq, "\n"))
    qual_line = " ".join(str(q) for q in qual)
    qualstring = "".join((name_line, qual_line, "\n"))
    return seqstring, qualstring


def format_as_fastq(name, seq, qual):
    qual_line = "".join(chr(q + 33) for q in qual)
    seqstring = "".join(("@", name, "\n", seq, "\n+\n", qual_line, "\n"))
    return seqstring


def get_read_data(data):
    """Given the data for one read it returns 2 strs with the fasta seq
    and fasta qual."""
    # seq and qual
    if config["mix_case"]:
        seq = sequence_case(data)
        qual = data["quality_scores"]
    else:
        seq = data["bases"]
        qual = data["quality_scores"]

    return seq, qual


def extract_read_info(data, fname):
    """Given the data for one read it returns 3 strs with the fasta seq, fasta
    qual and xml ancillary data."""
    seq, qual = get_read_data(data)
    seqstring, qualstring = format_as_fasta(data["name"], seq, qual)
    xmlstring = create_xml_for_unpaired_read(data, fname)
    return seqstring, qualstring, xmlstring


def write_sequence(name, seq, qual, seq_fh, qual_fh):
    """Write sequence and quality FASTA and FASTA qual filehandles
    (or into FASTQ and XML)
    if sequence length is 0, don't write"""
    if len(seq) == 0:
        return

    if qual_fh is None:
        seq_fh.write(format_as_fastq(name, seq, qual))
    else:
        seqstring, qualstring = format_as_fasta(name, seq, qual)
        seq_fh.write(seqstring)
        qual_fh.write(qualstring)
    return


def write_unpaired_read(data, sff_fh, seq_fh, qual_fh, xml_fh):
    """Writes an unpaired read into FASTA, FASTA qual and XML filehandles
    (or into FASTQ and XML)
    if sequence length is 0, don't write"""
    seq, qual = get_read_data(data)
    if len(seq) == 0:
        return

    write_sequence(data["name"], seq, qual, seq_fh, qual_fh)

    anci = create_xml_for_unpaired_read(data, sff_fh.name)
    if anci is not None:
        xml_fh.write(anci)
    return


def reverse_complement(seq):
    """Returns the reverse complement of a DNA sequence as string"""
    compdict = {
        "a": "t",
        "c": "g",
        "g": "c",
        "t": "a",
        "u": "t",
        "m": "k",
        "r": "y",
        "w": "w",
        "s": "s",
        "y": "r",
        "k": "m",
        "v": "b",
        "h": "d",
        "d": "h",
        "b": "v",
        "x": "x",
        "n": "n",
        "A": "T",
        "C": "G",
        "G": "C",
        "T": "A",
        "U": "T",
        "M": "K",
        "R": "Y",
        "W": "W",
        "S": "S",
        "Y": "R",
        "K": "M",
        "V": "B",
        "H": "D",
        "D": "H",
        "B": "V",
        "X": "X",
        "N": "N",
        "*": "*",
    }

    complseq = "".join(compdict[base] for base in seq)
    # python hack to reverse a list/string/etc
    complseq = complseq[::-1]
    return complseq


def mask_sequence(seq, maskchar, fpos, tpos):
    """Given a sequence, mask it with maskchar starting at fpos (including) and
    ending at tpos (excluding)
    """
    if len(maskchar) > 1:
        raise RuntimeError("Internal error: more than one character given to mask_sequence")
    if fpos < 0:
        fpos = 0
    if tpos > len(seq):
        tpos = len(seq)

    newseq = "".join((seq[:fpos], maskchar * (tpos - fpos), seq[tpos:]))

    return newseq


def fragment_sequences(sequence, qualities, splitchar):
    """Works like split() on strings, except it does this on a sequence
    and the corresponding list with quality values.
    Returns a tuple for each fragment, each sublist has the fragment
    sequence as first and the fragment qualities as second elemnt"""
    # this is slow (due to zip and list appends... use an iterator over
    #  the sequence find find variations and splices on seq and qual

    if len(sequence) != len(qualities):
        print(sequence, qualities)
        raise RuntimeError("Internal error: length of sequence and qualities don't match???")

    retlist = []
    if len(sequence) == 0:
        return retlist

    actseq = []
    actqual = []
    if sequence[0] != splitchar:
        inseq = True
    else:
        inseq = False
    for char, qual in zip(sequence, qualities):
        if inseq:
            if char != splitchar:
                actseq.append(char)
                actqual.append(qual)
            else:
                retlist.append(("".join(actseq), actqual))
                actseq = []
                actqual = []
                inseq = False
        else:
            if char != splitchar:
                inseq = True
                actseq.append(char)
                actqual.append(qual)

    if inseq and len(actseq):
        retlist.append(("".join(actseq), actqual))

    return retlist


def calc_subseq_boundaries(maskedseq, maskchar):
    """E.g.:
    ........xxxxxxxx..........xxxxxxxxxxxxxxxxxxxxx.........
    to
      (0,8),(8,16),(16,26),(26,47),(47,56)
    """
    blist = []
    if len(maskedseq) == 0:
        return blist

    inmask = True
    if maskedseq[0] != maskchar:
        inmask = False

    start = 0
    for spos in range(len(maskedseq)):
        if inmask and maskedseq[spos] != maskchar:
            blist.append([start, spos])
            start = spos
            inmask = False
        elif not inmask and maskedseq[spos] == maskchar:
            blist.append([start, spos])
            start = spos
            inmask = True

    blist.append([start, spos + 1])

    return blist


def correct_for_smallhits(maskedseq, maskchar, linkername):
    """If partial hits were found, take preventive measure: grow
     the masked areas by 20 bases in each direction
    Returns either unchanged "maskedseq" or a new sequence
     with some more characters masked.
    """
    if len(maskedseq) == 0:
        return maskedseq

    growl = 40
    growl2 = growl / 2

    boundaries = calc_subseq_boundaries(maskedseq, maskchar)

    foundpartial = False
    for bounds in boundaries:
        left, right = bounds
        if left != 0 and right != len(maskedseq):
            if maskedseq[left] == maskchar:
                # allow 10% discrepancy
                #    -linkerlengths[linkername]/10
                # that's a kind of safety net if there are slight sequencing
                #  errors in the linker itself
                if right - left < linkerlengths[linkername] - linkerlengths[linkername] / 10:
                    foundpartial = True

    if not foundpartial:
        return maskedseq

    # grow
    newseq = ""
    for bounds in boundaries:
        left, right = bounds
        if maskedseq[left] == maskchar:
            newseq += maskedseq[left:right]
        else:
            clearstart = 0
            if left > 0:
                clearstart = left + growl2
            clearstop = len(maskedseq)
            if right < len(maskedseq):
                clearstop = right - growl2

            if clearstop <= clearstart:
                newseq += maskchar * (right - left)
            else:
                if clearstart != left:
                    newseq += maskchar * growl2
                newseq += maskedseq[clearstart:clearstop]
                if clearstop != right:
                    newseq += maskchar * growl2

    return newseq


def split_paired_end(data, sff_fh, seq_fh, qual_fh, xml_fh):
    """Splits a paired end read and writes sequences into FASTA, FASTA qual
    and XML traceinfo file. Returns the number of sequences created.

    As the linker sequence may be anywhere in the read, including the ends
    and overlapping with bad quality sequence, we need to perform some
    computing and eventually set new clip points.

    If the resulting split yields only one sequence (because linker
    was not present or overlapping with left or right clip), only one
    sequence will be written with ".fn" appended to the name.

    If the read can be split, two reads will be written. The side left of
    the linker will be named ".r" and will be written in reverse complement
    into the file to conform with what approximately all assemblers expect
    when reading paired-end data: reads in forward direction in file. The side
    right of the linker will be named ".f"

    If SSAHA found partial linker (linker sequences < length of linker),
    the sequences will get a "_pl" furthermore be cut back thoroughly.

    If SSAHA found multiple occurences of the linker, the names will get an
    additional "_mlc" within the name to show that there was "multiple
    linker contamination".

    For multiple or partial linker, the "good" parts of the reads are
    stored with a ".part<number>" name, additionally they will not get
    template information in the XML
    """
    maskchar = "#"

    numseqs = 0
    readname = data["name"]
    readlen = data["number_of_bases"]

    leftclip, rightclip = return_merged_clips(data)
    seq, qual = get_read_data(data)

    maskedseq = seq
    if leftclip > 0:
        maskedseq = mask_sequence(maskedseq, maskchar, 0, leftclip - 1)
    if rightclip < len(maskedseq):
        maskedseq = mask_sequence(maskedseq, maskchar, rightclip, len(maskedseq))

    leftclip, rightclip = return_merged_clips(data)
    readlen = data["number_of_bases"]

    for match in ssahapematches[data["name"]]:
        int(match[0])
        linkername = match[2]
        leftreadhit = int(match[3])
        rightreadhit = int(match[4])

        maskedseq = mask_sequence(maskedseq, maskchar, leftreadhit - 1, rightreadhit)

    correctedseq = correct_for_smallhits(maskedseq, maskchar, linkername)

    if len(maskedseq) != len(correctedseq):
        raise RuntimeError("Internal error: maskedseq != correctedseq")

    partialhits = False
    if correctedseq != maskedseq:
        partialhits = True
        readname += "_pl"
        maskedseq = correctedseq

    fragments = fragment_sequences(maskedseq, qual, maskchar)

    mlcflag = False

    if len(fragments) > 2:
        mlcflag = True
        readname += "_mlc"

    # print fragments
    if mlcflag or partialhits:
        fragcounter = 1
        readname += ".part"
        for frag in fragments:
            actseq = frag[0]
            if len(actseq) >= 20:
                actqual = frag[1]
                oname = readname + str(fragcounter)
                write_sequence(oname, actseq, actqual, seq_fh, qual_fh)
                to_print = [create_basic_xml_info(oname, sff_fh.name)]
                # No clipping in XML ... the multiple and partial fragments
                #  are clipped "hard"
                # No template ID and trace_end: we don't know the
                #  orientation of the frahments. Even if it were
                #  only two, the fact we had multiple linkers
                #  says something went wrong, so simply do not
                #  write any paired-end information for all these fragments
                to_print.append("    </trace>\n")
                xml_fh.write("".join(to_print))
                numseqs += 1
                fragcounter += 1
    else:
        if len(fragments) > 2:
            raise RuntimeError(
                "Unexpected: more than two fragments detected in " + readname + ". please contact the authors."
            )
        # nothing will happen for 0 fragments
        if len(fragments) == 1:
            boundaries = calc_subseq_boundaries(maskedseq, maskchar)
            if len(boundaries) < 1 or len(boundaries) > 3:
                raise RuntimeError("Unexpected case: ", str(len(boundaries)), "boundaries for 1 fragment of ", readname)
            if len(boundaries) == 3:
                # case: mask char on both sides of sequence
                data["clip_adapter_left"] = boundaries[0][1]
                data["clip_adapter_right"] = boundaries[2][0]
            elif len(boundaries) == 2:
                # case: mask char left or right of sequence
                if maskedseq[0] == maskchar:
                    # case: mask char left
                    data["clip_adapter_left"] = boundaries[0][1]
                else:
                    # case: mask char right
                    data["clip_adapter_right"] = boundaries[1][0]
            data["name"] = data["name"] + ".fn"
            write_unpaired_read(data, sff_fh, seq_fh, qual_fh, xml_fh)
            numseqs = 1
        elif len(fragments) == 2:
            oname = readname + ".r"
            seq, qual = get_read_data(data)

            startsearch = False
            for spos in range(len(maskedseq)):
                if maskedseq[spos] != maskchar:
                    startsearch = True
                else:
                    if startsearch:
                        break

            lseq = seq[:spos]
            actseq = reverse_complement(lseq)
            lreadlen = len(actseq)
            lqual = qual[:spos]
            # python hack to reverse a list/string/etc
            lqual = lqual[::-1]

            write_sequence(oname, actseq, lqual, seq_fh, qual_fh)

            to_print = [create_basic_xml_info(oname, sff_fh.name)]
            to_print.append(
                create_clip_xml_info(
                    lreadlen, 0, lreadlen + 1 - data["clip_adapter_left"], 0, lreadlen + 1 - data["clip_qual_left"]
                )
            )
            to_print.append("        <template_id>")
            to_print.append(readname)
            to_print.append("</template_id>\n")
            to_print.append("        <trace_end>r</trace_end>\n")
            to_print.append("    </trace>\n")
            xml_fh.write("".join(to_print))

            oname = readname + ".f"
            startsearch = False
            for spos in range(len(maskedseq) - 1, -1, -1):
                if maskedseq[spos] != maskchar:
                    startsearch = True
                else:
                    if startsearch:
                        break

            actseq = seq[spos + 1 :]
            actqual = qual[spos + 1 :]

            write_sequence(oname, actseq, actqual, seq_fh, qual_fh)

            rreadlen = len(actseq)
            to_print = [create_basic_xml_info(oname, sff_fh.name)]
            to_print.append(
                create_clip_xml_info(
                    rreadlen,
                    0,
                    rreadlen - (readlen - data["clip_adapter_right"]),
                    0,
                    rreadlen - (readlen - data["clip_qual_right"]),
                )
            )
            to_print.append("        <template_id>")
            to_print.append(readname)
            to_print.append("</template_id>\n")
            to_print.append("        <trace_end>f</trace_end>\n")
            to_print.append("    </trace>\n")
            xml_fh.write("".join(to_print))
            numseqs = 2

    return numseqs


def extract_reads_from_sff(config, sff_files):
    """Given the configuration and the list of sff_files it writes the seqs,
    qualities and ancillary data into the output file(s).

    If file for paired-end linker was given, first extracts all sequences
    of an SFF and searches these against the linker(s) with SSAHA2 to
    create needed information to split reads.
    """
    if len(sff_files) == 0:
        raise RuntimeError("No SFF file given?")

    # we go through all input files
    for sff_file in sff_files:
        if not os.path.getsize(sff_file):
            raise RuntimeError("Empty file? : " + sff_file)
        fh = open(sff_file)
        fh.close()

    openmode = "w"
    if config["append"]:
        openmode = "a"

    seq_fh = open(config["seq_fname"], openmode)
    xml_fh = open(config["xml_fname"], openmode)
    if config["want_fastq"]:
        qual_fh = None
        try:
            os.remove(config["qual_fname"])
        except Exception:
            pass
    else:
        qual_fh = open(config["qual_fname"], openmode)

    if not config["append"]:
        xml_fh.write('<?xml version="1.0"?>\n<trace_volume>\n')
    else:
        remove_last_xmltag_in_file(config["xml_fname"], "trace_volume")

    # we go through all input files
    for sff_file in sff_files:
        ssahapematches.clear()

        seqcheckstore = []

        debug = 0

        if not debug and config["pelinker_fname"]:
            sys.stdout.flush()

            if 0:
                # for debugging
                pid = os.getpid()
                tmpfasta_fname = "sffe.tmp." + str(pid) + ".fasta"
                tmpfasta_fh = open(tmpfasta_fname, "w")
            else:
                tmpfasta_fh = tempfile.NamedTemporaryFile(prefix="sffeseqs_", suffix=".fasta")

            sff_fh = open(sff_file, "rb")
            header_data = read_header(fileh=sff_fh)
            for seq_data in sequences(fileh=sff_fh, header=header_data):
                seq, qual = get_read_data(seq_data)
                seqstring, qualstring = format_as_fasta(seq_data["name"], seq, qual)
                tmpfasta_fh.write(seqstring)
            tmpfasta_fh.seek(0)

            if 0:
                # for debugging
                tmpssaha_fname = "sffe.tmp." + str(pid) + ".ssaha2"
                tmpssaha_fh = open(tmpssaha_fname, "w+")
            else:
                tmpssaha_fh = tempfile.NamedTemporaryFile(prefix="sffealig_", suffix=".ssaha2")

            launch_ssaha(config["pelinker_fname"], tmpfasta_fh.name, tmpssaha_fh)
            tmpfasta_fh.close()

            tmpssaha_fh.seek(0)
            read_ssaha_data(tmpssaha_fh)
            tmpssaha_fh.close()

        if debug:
            tmpssaha_fh = open("sffe.tmp.10634.ssaha2")
            read_ssaha_data(tmpssaha_fh)

        sys.stdout.flush()
        sff_fh = open(sff_file, "rb")
        header_data = read_header(fileh=sff_fh)

        # now convert all reads
        nseqs_sff = 0
        nseqs_out = 0
        for seq_data in sequences(fileh=sff_fh, header=header_data):
            nseqs_sff += 1

            seq, qual = clip_read(seq_data)
            seqcheckstore.append(seq[0:50])

            if seq_data["name"] in ssahapematches:
                nseqs_out += split_paired_end(seq_data, sff_fh, seq_fh, qual_fh, xml_fh)
            else:
                if config["pelinker_fname"]:
                    seq_data["name"] = seq_data["name"] + ".fn"
                write_unpaired_read(seq_data, sff_fh, seq_fh, qual_fh, xml_fh)
                nseqs_out += 1
        sff_fh.close()

        check_for_dubious_startseq(seqcheckstore, sff_file, seq_data)
        seqcheckstore = []

    xml_fh.write("</trace_volume>\n")

    xml_fh.close()
    seq_fh.close()
    if qual_fh is not None:
        qual_fh.close()

    return


def check_for_dubious_startseq(seqcheckstore, sffname, seqdata):
    global stern_warning

    foundproblem = ""
    for checklen in range(1, len(seqcheckstore[0])):
        foundinloop = False
        seqdict = {}
        for seq in seqcheckstore:
            shortseq = seq[0:checklen]
            if shortseq in seqdict:
                seqdict[shortseq] += 1
            else:
                seqdict[shortseq] = 1

        for shortseq, count in seqdict.items():
            if float(count) / len(seqcheckstore) >= 0.5:
                foundinloop = True
                stern_warning = True
                foundproblem = "\n" + "*" * 80
                foundproblem += "\nWARNING: "
                foundproblem += "weird sequences in file " + sffname + "\n\n"
                foundproblem += "After applying left clips, " + str(count) + " sequences (="
                foundproblem += "%.0f" % (100.0 * float(count) / len(seqcheckstore))
                foundproblem += "%) start with these bases:\n" + shortseq
                foundproblem += "\n\nThis does not look sane.\n\n"
                foundproblem += "Countermeasures you *probably* must take:\n"
                foundproblem += "1) Make your sequence provider aware of that problem and ask whether this can be\n    corrected in the SFF.\n"
                foundproblem += "2) If you decide that this is not normal and your sequence provider does not\n    react, use the --min_left_clip of sff_extract.\n"
                left, right = return_merged_clips(seqdata)
                foundproblem += (
                    "    (Probably '--min_left_clip="
                    + str(left + len(shortseq))
                    + "' but you should cross-check that)\n"
                )
                foundproblem += "*" * 80 + "\n"
        if not foundinloop:
            break
    if len(foundproblem):
        print(foundproblem)


def parse_extra_info(info):
    """It parses the information that will go in the xml file.

    There are two formats accepted for the extra information:
    key1:value1, key2:value2
    or:
    file1.sff{key1:value1, key2:value2};file2.sff{key3:value3}
    """
    if not info:
        return info
    finfos = info.split(";")  # information for each file
    data_for_files = {}
    for finfo in finfos:
        # we split the file name from the rest
        items = finfo.split("{")
        if len(items) == 1:
            fname = fake_sff_name
            info = items[0]
        else:
            fname = items[0]
            info = items[1]
        # now we get each key,value pair in the info
        info = info.replace("}", "")
        data = {}
        for item in info.split(","):
            key, value = item.strip().split(":")
            key = key.strip()
            value = value.strip()
            data[key] = value
        data_for_files[fname] = data
    return data_for_files


def return_merged_clips(data):
    """It returns the left and right positions to clip."""

    def max(a, b):
        """It returns the max of the two given numbers.

        It won't take into account the zero values.
        """
        if not a and not b:
            return None
        if not a:
            return b
        if not b:
            return a
        if a >= b:
            return a
        else:
            return b

    def min(a, b):
        """It returns the min of the two given numbers.

        It won't take into account the zero values.
        """
        if not a and not b:
            return None
        if not a:
            return b
        if not b:
            return a
        if a <= b:
            return a
        else:
            return b

    left = max(data["clip_adapter_left"], data["clip_qual_left"])
    right = min(data["clip_adapter_right"], data["clip_qual_right"])
    # maybe both clips where zero
    if left is None:
        left = 1
    if right is None:
        right = data["number_of_bases"]
    return left, right


def sequence_case(data):
    """Given the data for one read it returns the seq with mixed case.

    The regions to be clipped will be lower case and the rest upper case.
    """
    left, right = return_merged_clips(data)
    seq = data["bases"]
    if left >= right:
        new_seq = seq.lower()
    else:
        new_seq = "".join((seq[: left - 1].lower(), seq[left - 1 : right], seq[right:].lower()))

    return new_seq


def clip_read(data):
    """Given the data for one read it returns clipped seq and qual."""
    qual = data["quality_scores"]
    left, right = return_merged_clips(data)
    seq = data["bases"]
    qual = data["quality_scores"]
    new_seq = seq[left - 1 : right]
    new_qual = qual[left - 1 : right]

    return new_seq, new_qual


def tests_for_ssaha():
    """Tests whether SSAHA2 can be successfully called."""
    try:
        print("Testing whether SSAHA2 is installed and can be launched ... ", end=" ")
        sys.stdout.flush()
        fh = open("/dev/null", "w")
        subprocess.call(["ssaha2"], stdout=fh)
        fh.close()
        print("ok.")
    except Exception:
        print("nope? Uh oh ...\n\n")
        raise RuntimeError("Could not launch ssaha2. Have you installed it? Is it in your path?")


def load_linker_sequences(linker_fname):
    """Loads all linker sequences into memory, storing only the length
    of each linker."""
    if not os.path.getsize(linker_fname):
        raise RuntimeError("File empty? '" + linker_fname + "'")
    fh = open(linker_fname)
    linkerseqs = read_fasta(fh)
    if len(linkerseqs) == 0:
        raise RuntimeError(linker_fname + ": no sequence found?")
    for i in linkerseqs:
        if i.name in linkerlengths:
            raise RuntimeError(linker_fname + ": sequence '" + i.name + "' present multiple times. Aborting.")
        linkerlengths[i.name] = len(i.sequence)
    fh.close()


def launch_ssaha(linker_fname, query_fname, output_fh):
    """Launches SSAHA2 on the linker and query file, string SSAHA2 output
    into the output filehandle"""
    tests_for_ssaha()

    try:
        print("Searching linker sequences with SSAHA2 (this may take a while) ... ", end=" ")
        sys.stdout.flush()
        retcode = subprocess.call(
            ["ssaha2", "-output", "ssaha2", "-solexa", "-kmer", "4", "-skip", "1", linker_fname, query_fname],
            stdout=output_fh,
        )
        if retcode:
            raise RuntimeError("Ups.")
        else:
            print("ok.")
    except Exception:
        print("\n")
        raise RuntimeError("An error occurred during the SSAHA2 execution, aborting.")


def read_ssaha_data(ssahadata_fh):
    """Given file handle, reads file generated with SSAHA2 (with default
    output format) and stores all matches as list ssahapematches
    (ssaha paired-end matches) dictionary"""
    print("Parsing SSAHA2 result file ... ", end=" ")
    sys.stdout.flush()

    for line in ssahadata_fh:
        if line.startswith("ALIGNMENT"):
            ml = line.split()
            if len(ml) != 12:
                print("\n", line, end=" ")
                raise RuntimeError(
                    "Expected 12 elements in the SSAHA2 line with ALIGMENT keyword, but found " + str(len(ml))
                )
            if ml[2] not in ssahapematches:
                ssahapematches[ml[2]] = []
            if ml[8] == "F":
                # store everything except the first element (output
                #  format name (ALIGNMENT)) and the last element
                #  (length)
                ssahapematches[ml[2]].append(ml[1:-1])
            else:
                ml[4], ml[5] = ml[5], ml[4]
                ssahapematches[ml[2]].append(ml[1:-1])

    print("done.")


##########################################################################
#
# BaCh: This block was shamelessly copied from
#  http://python.genedrift.org/2007/07/04/reading-fasta-files-conclusion/
# and then subsequently modified to read fasta correctly
# It's still not fool proof, but should be good enough
#
##########################################################################


class Fasta(object):
    def __init__(self, name, sequence):
        self.name = name
        self.sequence = sequence


def read_fasta(file):
    items = []
    aninstance = Fasta("", "")
    linenum = 0
    for line in file:
        linenum += 1
        if line.startswith(">"):
            if len(aninstance.sequence):
                items.append(aninstance)
                aninstance = Fasta("", "")
            # name == all characters until the first whitespace
            #  (split()[0]) but without the starting ">" ([1:])
            aninstance.name = line.split()[0][1:]
            aninstance.sequence = ""
            if len(aninstance.name) == 0:
                raise RuntimeError(file.name + ": no name in line " + str(linenum) + "?")

        else:
            if len(aninstance.name) == 0:
                raise RuntimeError(file.name + ": no sequence header at line " + str(linenum) + "?")
            aninstance.sequence += line.strip()

    if len(aninstance.name) and len(aninstance.sequence):
        items.append(aninstance)

    return items


##########################################################################


def version_string():
    return "sff_extract " + __version__


def read_config():
    """It reads the configuration options from the command line arguments and
    it returns a dict with them."""
    from optparse import (
        OptionGroup,
        OptionParser,
    )

    usage = "usage: %prog [options] sff1 sff2 ..."
    desc = (
        "Extract sequences from 454 SFF files into FASTA, FASTA quality"
        " and XML traceinfo format. When a paired-end linker sequence"
        " is given (-l), use SSAHA2 to scan the sequences for the linker,"
        " then split the sequences, removing the linker."
    )
    parser = OptionParser(usage=usage, version=version_string(), description=desc)
    parser.add_option(
        "-a", "--append", action="store_true", dest="append", help="append output to existing files", default=False
    )
    parser.add_option("-i", "--xml_info", dest="xml_info", help="extra info to write in the xml file")
    parser.add_option(
        "-l", "--linker_file", dest="pelinker_fname", help="FASTA file with paired-end linker sequences", metavar="FILE"
    )

    group = OptionGroup(parser, "File name options", "")
    group.add_option(
        "-c",
        "--clip",
        action="store_true",
        dest="clip",
        help="clip (completely remove) ends with low qual and/or adaptor sequence",
        default=False,
    )
    group.add_option(
        "-u",
        "--upper_case",
        action="store_false",
        dest="mix_case",
        help="all bases in upper case, including clipped ends",
        default=True,
    )
    group.add_option(
        "",
        "--min_left_clip",
        dest="min_leftclip",
        metavar="INTEGER",
        type="int",
        help="if the left clip coming from the SFF is smaller than this value, override it",
        default=0,
    )
    group.add_option(
        "-Q",
        "--fastq",
        action="store_true",
        dest="want_fastq",
        help="store as FASTQ file instead of FASTA + FASTA quality file",
        default=False,
    )
    parser.add_option_group(group)

    group = OptionGroup(parser, "File name options", "")
    group.add_option("-o", "--out_basename", dest="basename", help="base name for all output files")
    group.add_option("-s", "--seq_file", dest="seq_fname", help="output sequence file name", metavar="FILE")
    group.add_option("-q", "--qual_file", dest="qual_fname", help="output quality file name", metavar="FILE")
    group.add_option("-x", "--xml_file", dest="xml_fname", help="output ancillary xml file name", metavar="FILE")
    parser.add_option_group(group)

    # default fnames
    # is there an sff file?
    basename = "reads"
    if sys.argv[-1][-4:].lower() == ".sff":
        basename = sys.argv[-1][:-4]
    def_seq_fname = basename + ".fasta"
    def_qual_fname = basename + ".fasta.qual"
    def_xml_fname = basename + ".xml"
    def_pelinker_fname = ""
    parser.set_defaults(seq_fname=def_seq_fname)
    parser.set_defaults(qual_fname=def_qual_fname)
    parser.set_defaults(xml_fname=def_xml_fname)
    parser.set_defaults(pelinker_fname=def_pelinker_fname)

    # we parse the cmd line
    (options, args) = parser.parse_args()

    # we put the result in a dict
    global config
    config = {}
    for property in dir(options):
        if property[0] == "_" or property in ("ensure_value", "read_file", "read_module"):
            continue
        config[property] = getattr(options, property)

    if config["basename"] is None:
        config["basename"] = basename

    # if we have not set a file name with -s, -q or -x we set the basename
    # based file name
    if config["want_fastq"]:
        config["qual_fname"] = ""
        if config["seq_fname"] == def_seq_fname:
            config["seq_fname"] = config["basename"] + ".fastq"
    else:
        if config["seq_fname"] == def_seq_fname:
            config["seq_fname"] = config["basename"] + ".fasta"
        if config["qual_fname"] == def_qual_fname:
            config["qual_fname"] = config["basename"] + ".fasta.qual"

    if config["xml_fname"] == def_xml_fname:
        config["xml_fname"] = config["basename"] + ".xml"

    # we parse the extra info for the xml file
    config["xml_info"] = parse_extra_info(config["xml_info"])
    return config, args


def testsome():
    sys.exit()
    return


def main():
    argv = sys.argv
    if len(argv) == 1:
        sys.argv.append("-h")
        read_config()
        sys.exit()
    try:
        config, args = read_config()

        if config["pelinker_fname"]:
            load_linker_sequences(config["pelinker_fname"])
        if len(args) == 0:
            raise RuntimeError("No SFF file given?")
        extract_reads_from_sff(config, args)
    except (OSError, IOError, RuntimeError) as errval:  # noqa: B014 Python 2.7
        print(errval)
        return 1

    if stern_warning:
        return 1

    return 0


if __name__ == "__main__":
    sys.exit(main())



import logging
import os
from typing import (
    Any,
    cast,
    Dict,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
)

from pydantic import (
    BaseModel,
    UUID4,
)
from typing_extensions import TypedDict

from galaxy.exceptions import (
    InconsistentDatabase,
    ObjectNotFound,
    RequestParameterInvalidException,
    RequestParameterMissingException,
)
from galaxy.managers.context import ProvidesUserContext
from galaxy.model import (
    CONFIGURATION_TEMPLATE_CONFIGURATION_VARIABLES_TYPE,
    HasConfigEnvironment,
    HasConfigSecrets,
    HasConfigTemplate,
    User,
    UsesTemplatesAppConfig,
)
from galaxy.model.scoped_session import galaxy_scoped_session
from galaxy.security.vault import (
    UserVaultWrapper,
    Vault,
)
from galaxy.util import config_templates  # defer get_token_from_refresh_raw access for monkey patching
from galaxy.util.config_templates import (
    EnvironmentDict,
    find_template_by,
    ImplicitConfigurationParameters,
    OAuth2ClientPair,
    OAuth2Configuration,
    PluginAspectStatus,
    secrets_as_dict,
    SecretsDict,
    Template,
    TemplateEnvironmentEntry,
    TemplateEnvironmentSecret,
    TemplateEnvironmentVariable,
    TemplateReference,
    TemplateVariableValueType,
    validate_no_extra_variables_defined,
    validate_specified_datatypes_variables,
)
from galaxy.work.context import SessionRequestContext

log = logging.getLogger(__name__)

SuppliedVariables = Dict[str, TemplateVariableValueType]
SuppliedSecrets = Dict[str, str]


class CreateInstancePayload(BaseModel):
    name: str
    description: Optional[str] = None
    template_id: str
    template_version: int
    variables: SuppliedVariables
    secrets: SuppliedSecrets
    uuid: Optional[UUID4] = None


class UpdateInstancePayload(BaseModel):
    name: Optional[str] = None
    description: Optional[str] = None
    variables: Optional[SuppliedVariables] = None
    hidden: Optional[bool] = None
    active: Optional[bool] = None


class UpdateInstanceSecretPayload(BaseModel):
    secret_name: str
    secret_value: str


class UpgradeInstancePayload(BaseModel):
    template_version: int
    variables: SuppliedVariables
    secrets: SuppliedSecrets


class TestUpdateInstancePayload(BaseModel):
    variables: Optional[SuppliedVariables] = None


class TestUpgradeInstancePayload(BaseModel):
    template_version: int
    variables: SuppliedVariables
    secrets: SuppliedSecrets


class UpgradeTestTarget:
    instance: HasConfigTemplate
    payload: TestUpgradeInstancePayload

    def __init__(self, instance: HasConfigTemplate, payload: TestUpgradeInstancePayload):
        self.instance = instance
        self.payload = payload


class UpdateTestTarget:
    instance: HasConfigTemplate
    payload: TestUpdateInstancePayload

    def __init__(self, instance: HasConfigTemplate, payload: TestUpdateInstancePayload):
        self.instance = instance
        self.payload = payload


class CreateTestTarget:
    payload: CreateInstancePayload
    instance_class: Type[HasConfigSecrets]

    def __init__(self, payload: CreateInstancePayload, instance_class: Type[HasConfigSecrets]):
        self.payload = payload
        self.instance_class = instance_class


ModifyInstancePayload = Union[UpdateInstanceSecretPayload, UpgradeInstancePayload, UpdateInstancePayload]
TestModifyInstancePayload = Union[TestUpgradeInstancePayload, TestUpdateInstancePayload]
CanTestPluginStatus = Union[HasConfigTemplate, CreateTestTarget, UpgradeTestTarget, UpdateTestTarget]


def recover_secrets(
    user_object_store: HasConfigSecrets, vault: Union[UserVaultWrapper, Vault], app_config: UsesTemplatesAppConfig
) -> SecretsDict:
    if isinstance(vault, UserVaultWrapper):
        user_vault = vault
    else:
        user: User = user_object_store.user
        user_vault = UserVaultWrapper(vault, user)
    secrets: SecretsDict = {}
    # now we could recover the list of secrets to fetch from...
    # ones recorded as written in the persisted object, the ones
    # expected in the catalog, or the ones expected in the definition
    # persisted.
    persisted_secret_names = user_object_store.template_secrets or []
    for secret in persisted_secret_names:
        vault_key = user_object_store.vault_key(secret, app_config)
        secret_value = user_vault.read_secret(vault_key)
        if secret_value is not None:
            secrets[secret] = secret_value
    return secrets


class TemplateParameters(TypedDict):
    secrets: SuppliedSecrets
    variables: SuppliedVariables
    environment: EnvironmentDict
    user_details: Dict[str, Any]
    implicit: Optional[ImplicitConfigurationParameters]


class TemplateServerConfiguration:
    """Extra configuration relevant to a given template.

    For state defined in the server but not defined in the template JSON/YAML directly.
    Currently this only contains optional oauth2 client credentials and oauth2 configuration
    information (provider URLs and provider specific configuration for the oauth2 flow).
    """

    oauth2_client_pair: Optional[OAuth2ClientPair]
    oauth2_configuration: Optional[OAuth2Configuration]

    def __init__(
        self,
        oauth2_client_pair: Optional[OAuth2ClientPair] = None,
        oauth2_configuration: Optional[OAuth2Configuration] = None,
        oauth2_scope: Optional[str] = None,
    ):
        self.oauth2_client_pair = oauth2_client_pair
        self.oauth2_configuration = oauth2_configuration
        self.oauth2_scope = oauth2_scope

    @property
    def uses_oauth2(self):
        return self.oauth2_configuration is not None


def prepare_template_parameters_for_testing(
    trans: ProvidesUserContext,
    template: Template,
    template_server_configuration: TemplateServerConfiguration,
    target: CanTestPluginStatus,
    vault: Vault,
    app_config: UsesTemplatesAppConfig,
) -> TemplateParameters:
    secrets = _secrets_for_plugin_status_test(target, trans.user_vault, app_config)
    variables = _variables_for_plugin_status_test(target)
    environment = prepare_environment_from_root(template.environment, vault, app_config)
    user_details = trans.user.config_template_details()
    implicit = implicit_parameters_for_testing(
        trans,
        template_server_configuration,
        target,
        app_config,
    )
    return TemplateParameters(
        {
            "secrets": secrets,
            "variables": variables,
            "environment": environment,
            "user_details": user_details,
            "implicit": implicit,
        }
    )


def _variables_for_plugin_status_test(target: CanTestPluginStatus) -> SuppliedVariables:
    if isinstance(target, CreateTestTarget):
        return target.payload.variables
    elif isinstance(target, UpgradeTestTarget) or isinstance(target, UpdateTestTarget):
        if target.instance.template_variables:
            variables = target.instance.template_variables.copy()
        else:
            variables = {}
        new_variables = target.payload.variables or {}
        for new_variable in new_variables:
            variables[new_variable] = new_variables[new_variable]
        return new_variables
    else:
        return target.template_variables or {}


def _secrets_for_plugin_status_test(
    target: CanTestPluginStatus, user_vault: UserVaultWrapper, app_config: UsesTemplatesAppConfig
) -> SecretsDict:
    if isinstance(target, CreateTestTarget):
        payload = target.payload
        secrets = payload.secrets.copy()
        return secrets
    elif isinstance(target, UpgradeTestTarget):
        secrets = recover_secrets(target.instance, user_vault, app_config)
        new_secrets = target.payload.secrets or {}
        for new_secret in new_secrets:
            secrets[new_secret] = new_secrets[new_secret]
        return secrets
    elif isinstance(target, UpdateTestTarget):
        secrets = recover_secrets(target.instance, user_vault, app_config)
        return secrets
    else:
        secrets = recover_secrets(target, user_vault, app_config)
        return secrets


def to_template_reference(persisted_instance: HasConfigTemplate) -> TemplateReference:
    # for mypy convert Mapped[X] -> X
    return cast(TemplateReference, persisted_instance)


def prepare_environment(
    configuration_template: HasConfigEnvironment, vault: Vault, app_config: UsesTemplatesAppConfig
) -> EnvironmentDict:
    return prepare_environment_from_root(configuration_template.template_environment.root, vault, app_config)


def prepare_environment_from_root(
    root: Optional[List[TemplateEnvironmentEntry]], vault: Vault, app_config: UsesTemplatesAppConfig
) -> EnvironmentDict:
    environment: EnvironmentDict = {}
    for environment_entry in root or []:
        e_type = environment_entry.type
        e_name = environment_entry.name
        if e_type == "secret":
            template_secret = cast(TemplateEnvironmentSecret, environment_entry)
            secret_value = vault.read_secret(template_secret.vault_key) or template_secret.default
            if secret_value:
                environment[e_name] = secret_value
        elif e_type == "variable":
            template_variable = cast(TemplateEnvironmentVariable, environment_entry)
            variable_value = os.environ.get(template_variable.variable)
            if variable_value is None:
                variable_value = template_variable.default
            if variable_value:
                environment[e_name] = variable_value
        else:
            raise Exception(f"Unknown environment entry type detected [{e_type}]")

    return environment


def update_template_instance(
    sa_session: galaxy_scoped_session,
    template_instance: HasConfigTemplate,
    payload: UpdateInstancePayload,
    template: Template,
):
    validate_specified_datatypes_variables(payload.variables or {}, template)
    validate_no_extra_variables_defined(payload.variables or {}, template)
    if payload.name is not None:
        template_instance.name = payload.name
    if payload.description is not None:
        template_instance.description = payload.description
    if payload.variables is not None:
        actual_variables = updated_template_variables(payload.variables, template_instance, template)
        template_instance.template_variables = actual_variables
    if payload.hidden is not None:
        template_instance.hidden = payload.hidden
    if payload.active is not None:
        if payload.active:
            # unhide and activate
            template_instance.active = True
            template_instance.hidden = False
        else:
            # deactivate and hide
            template_instance.active = False
            template_instance.hidden = True
    save_template_instance(sa_session, template_instance)
    return template_instance


def updated_template_variables(
    supplied_variables: SuppliedVariables, template_instance: HasConfigTemplate, template: Template
):
    old_variables = template_instance.template_variables or {}
    updated_variables: CONFIGURATION_TEMPLATE_CONFIGURATION_VARIABLES_TYPE = {}
    for variable in template.variables or []:
        variable_name = variable.name
        old_value = old_variables.get(variable_name)
        updated_value = supplied_variables.get(variable_name, old_value)
        if updated_value:
            updated_variables[variable_name] = updated_value
    return updated_variables


def purge_template_instance(
    trans: ProvidesUserContext, template_instance: HasConfigTemplate, app_config: UsesTemplatesAppConfig
):
    user_vault = trans.user_vault
    user_vault_key_prefix = template_instance.vault_id_prefix(app_config)
    # it would be nice to just list the secrets but none of them do that... so catch NotImplemented
    # and try with the secrets as we have record in the database of recording them.
    try:
        for secret in user_vault.list_secrets(user_vault_key_prefix):
            user_vault.delete_secret(secret)
    except NotImplementedError:
        for secret in template_instance.template_secrets or []:
            user_vault.delete_secret(f"{user_vault_key_prefix}/{secret}")
    template_instance.active = False
    template_instance.purged = True
    save_template_instance(trans.sa_session, template_instance)


def update_instance_secret(
    trans: ProvidesUserContext,
    template_instance: HasConfigTemplate,
    template: Template,
    payload: UpdateInstanceSecretPayload,
    app_config: UsesTemplatesAppConfig,
):
    template_secrets = secrets_as_dict(template.secrets or [])
    secret_name = payload.secret_name
    if secret_name not in template_secrets:
        raise RequestParameterInvalidException(f"Configuration template does not specify a secret named {secret_name}")

    user_vault = trans.user_vault
    key = template_instance.vault_key(payload.secret_name, app_config)
    user_vault.write_secret(key, payload.secret_value)


def upgrade_secrets(
    trans: ProvidesUserContext,
    template_instance: HasConfigTemplate,
    target_template: Template,
    payload: UpgradeInstancePayload,
    app_config: UsesTemplatesAppConfig,
):
    recorded_secrets = template_instance.template_secrets or []

    old_secrets = template_instance.template_secrets or []
    new_secrets = payload.secrets

    user_vault = trans.user_vault
    upgraded_template_secrets = []
    for secret in target_template.secrets or []:
        secret_name = secret.name
        upgraded_template_secrets.append(secret_name)
        if secret_name not in new_secrets and secret_name not in old_secrets:
            raise RequestParameterMissingException(f"secret {secret_name} not set in supplied request")
        if secret_name not in new_secrets:
            # keep old value
            continue

        secret_value = new_secrets[secret_name]
        key = template_instance.vault_key(secret_name, app_config)
        user_vault.write_secret(key, secret_value)
        if secret_name not in recorded_secrets:
            recorded_secrets.append(secret_name)

    secrets_to_delete: List[str] = []
    for recorded_secret in recorded_secrets:
        if recorded_secret not in upgraded_template_secrets:
            key = template_instance.vault_key(recorded_secret, app_config)
            log.info(f"deleting {key} from user vault")
            user_vault.delete_secret(key)
            secrets_to_delete.append(recorded_secret)

    for secret_to_delete in secrets_to_delete:
        recorded_secrets.remove(secret_to_delete)

    template_instance.template_secrets = recorded_secrets


def save_template_instance(sa_session: galaxy_scoped_session, template_instance: HasConfigTemplate):
    sa_session.add(template_instance)
    sa_session.flush([template_instance])
    sa_session.commit()


T = TypeVar("T", bound=Template, covariant=True)


def sort_templates(config, catalog: List[T], instance: HasConfigTemplate) -> List[T]:
    configured_template: Optional[T] = None
    try:
        configured_template = find_template_by(
            catalog, instance.template_id, instance.template_version, "config template"
        )
    except ObjectNotFound:
        if config.user_config_templates_use_saved_configuration == "never":
            raise
    stored_template: T = instance.template
    if config.user_config_templates_use_saved_configuration == "preferred" and configured_template:
        templates = [stored_template, configured_template]
    elif configured_template:
        templates = [configured_template, stored_template]
    else:
        templates = [stored_template]
    return templates


def implicit_parameters_for_testing(
    trans: ProvidesUserContext,
    template_server_configuration: TemplateServerConfiguration,
    target: CanTestPluginStatus,
    app_config: UsesTemplatesAppConfig,
) -> Optional[ImplicitConfigurationParameters]:
    implicit: ImplicitConfigurationParameters = {}
    if template_server_configuration.oauth2_configuration:
        refresh_token_key = None
        oauth2_refresh_token = None
        if isinstance(target, CreateTestTarget):
            if target.payload.uuid:
                refresh_token_key = target.instance_class.vault_key_from_uuid(
                    target.payload.uuid, "_oauth2_refresh_token", app_config
                )
        else:
            if isinstance(target, (UpgradeTestTarget, UpdateTestTarget)):
                instance = target.instance
            else:
                instance = target
            refresh_token_key = instance.vault_key_from_uuid(instance.uuid, "_oauth2_refresh_token", app_config)
        oauth2_refresh_token = trans.user_vault.read_secret(refresh_token_key)
        if not oauth2_refresh_token:
            raise InconsistentDatabase(
                f"Failed to recover oauth2 refresh token from vault at location {refresh_token_key}, Galaxy is in an inconsistent state and probably requires admin intervention"
            )
        _inject_oauth2_access_token(implicit, oauth2_refresh_token, template_server_configuration)

    return implicit


def implicit_parameters_for_instance(
    user_instance: HasConfigSecrets,
    template_server_configuration: TemplateServerConfiguration,
    vault: Vault,
    app_config: UsesTemplatesAppConfig,
) -> ImplicitConfigurationParameters:
    implicit: ImplicitConfigurationParameters = {}

    if template_server_configuration.oauth2_configuration:
        refresh_token_key = user_instance.__class__.vault_key_from_uuid(
            str(user_instance.uuid), "_oauth2_refresh_token", app_config
        )
        user_vault = UserVaultWrapper(vault, user_instance.user)
        oauth2_refresh_token = user_vault.read_secret(refresh_token_key)
        if not oauth2_refresh_token:
            raise Exception("null refresh token key read from user vault")
        _inject_oauth2_access_token(implicit, oauth2_refresh_token, template_server_configuration)

    return implicit


def _inject_oauth2_access_token(
    implicit: ImplicitConfigurationParameters,
    oauth2_refresh_token: str,
    template_server_configuration: TemplateServerConfiguration,
) -> None:
    oauth2_client_pair = template_server_configuration.oauth2_client_pair
    oauth2_configuration = template_server_configuration.oauth2_configuration
    assert oauth2_client_pair
    assert oauth2_configuration
    response = config_templates.get_token_from_refresh_raw(
        oauth2_refresh_token, oauth2_client_pair, oauth2_configuration
    )
    response.raise_for_status()
    implicit["oauth2_access_token"] = response.json()["access_token"]


def oauth2_refresh_token_status(
    template_server_configuration: TemplateServerConfiguration, exception: Optional[Exception]
) -> Optional[PluginAspectStatus]:
    if not template_server_configuration.uses_oauth2:
        # no oauth enabled, don't report a status associated with
        return None
    else:
        if not exception:
            return PluginAspectStatus(
                state="ok", message="OAuth2 secret refresh token generated an access token for this resource"
            )
        else:
            return PluginAspectStatus(
                state="not_ok",
                message=f"OAuth2 secret refresh token failed to generate an access token for this resource: {exception}",
            )


def oauth2_redirect_uri(trans: SessionRequestContext) -> str:
    galaxy_root = trans.request.url_path
    redirect_uri = f"{galaxy_root}oauth2_callback"
    return redirect_uri



from __future__ import print_function

from shutil import copyfile


def load_microbial_data(GALAXY_DATA_INDEX_DIR, sep="\t"):
    # FIXME: this function is duplicated in the DynamicOptions class.  It is used here only to
    # set data.name in exec_after_process().
    microbe_info = {}
    orgs = {}

    filename = "%s/microbial_data.loc" % GALAXY_DATA_INDEX_DIR
    for line in open(filename):
        line = line.rstrip("\r\n")
        if line and not line.startswith("#"):
            fields = line.split(sep)
            # read each line, if not enough fields, go to next line
            try:
                info_type = fields.pop(0)
                if info_type.upper() == "ORG":
                    # ORG     12521   Clostridium perfringens SM101   bacteria        Firmicutes      CP000312,CP000313,CP000314,CP000315     http://www.ncbi.nlm.nih.gov/entrez/query.fcgi?db=genomeprj&cmd=Retrieve&dopt=Overview&list_uids=12521
                    org_num = fields.pop(0)
                    name = fields.pop(0)
                    kingdom = fields.pop(0)
                    group = fields.pop(0)
                    chromosomes = fields.pop(0)
                    info_url = fields.pop(0)
                    link_site = fields.pop(0)
                    if org_num not in orgs:
                        orgs[org_num] = {}
                        orgs[org_num]["chrs"] = {}
                    orgs[org_num]["name"] = name
                    orgs[org_num]["kingdom"] = kingdom
                    orgs[org_num]["group"] = group
                    orgs[org_num]["chromosomes"] = chromosomes
                    orgs[org_num]["info_url"] = info_url
                    orgs[org_num]["link_site"] = link_site
                elif info_type.upper() == "CHR":
                    # CHR     12521   CP000315        Clostridium perfringens phage phiSM101, complete genome 38092   110684521       CP000315.1
                    org_num = fields.pop(0)
                    chr_acc = fields.pop(0)
                    name = fields.pop(0)
                    length = fields.pop(0)
                    gi = fields.pop(0)
                    gb = fields.pop(0)
                    info_url = fields.pop(0)
                    chr = {}
                    chr["name"] = name
                    chr["length"] = length
                    chr["gi"] = gi
                    chr["gb"] = gb
                    chr["info_url"] = info_url
                    if org_num not in orgs:
                        orgs[org_num] = {}
                        orgs[org_num]["chrs"] = {}
                    orgs[org_num]["chrs"][chr_acc] = chr
                elif info_type.upper() == "DATA":
                    # DATA    12521_12521_CDS 12521   CP000315        CDS     bed     /home/djb396/alignments/playground/bacteria/12521/CP000315.CDS.bed
                    uid = fields.pop(0)
                    org_num = fields.pop(0)
                    chr_acc = fields.pop(0)
                    feature = fields.pop(0)
                    filetype = fields.pop(0)
                    path = fields.pop(0)
                    data = {}
                    data["filetype"] = filetype
                    data["path"] = path
                    data["feature"] = feature

                    if org_num not in orgs:
                        orgs[org_num] = {}
                        orgs[org_num]["chrs"] = {}
                    if "data" not in orgs[org_num]["chrs"][chr_acc]:
                        orgs[org_num]["chrs"][chr_acc]["data"] = {}
                    orgs[org_num]["chrs"][chr_acc]["data"][uid] = data
                else:
                    continue
            except Exception:
                continue
    for org_num in orgs:
        org = orgs[org_num]
        if org["kingdom"] not in microbe_info:
            microbe_info[org["kingdom"]] = {}
        if org_num not in microbe_info[org["kingdom"]]:
            microbe_info[org["kingdom"]][org_num] = org
    return microbe_info


# post processing, set build for data and add additional data to history
def exec_after_process(app, inp_data, out_data, param_dict, tool, stdout, stderr):
    base_dataset = next(iter(out_data.values()))
    history = base_dataset.history
    if history is None:
        print("unknown history!")
        return
    kingdom = param_dict.get("kingdom", None)
    org = param_dict.get("org", None)

    # if not (kingdom or group or org):
    if not (kingdom or org):
        print("Parameters are not available.")

    GALAXY_DATA_INDEX_DIR = app.config.tool_data_path
    microbe_info = load_microbial_data(GALAXY_DATA_INDEX_DIR, sep="\t")
    split_stdout = stdout.split("\n")
    basic_name = ""
    for line in split_stdout:
        fields = line.split("\t")
        if fields[0] == "#File1":
            description = fields[1]
            chr = fields[2]
            dbkey = fields[3]
            file_type = fields[4]
            data = next(iter(out_data.values()))
            data.set_size()
            basic_name = data.name
            data.name = (
                data.name
                + " ("
                + microbe_info[kingdom][org]["chrs"][chr]["data"][description]["feature"]
                + " for "
                + microbe_info[kingdom][org]["name"]
                + ":"
                + chr
                + ")"
            )
            data.dbkey = dbkey
            data.info = data.name
            data = app.datatypes_registry.change_datatype(data, file_type)
            data.init_meta()
            data.set_peek()
            app.model.context.add(data)
            app.model.context.flush()
        elif fields[0] == "#NewFile":
            description = fields[1]
            chr = fields[2]
            dbkey = fields[3]
            filepath = fields[4]
            file_type = fields[5]
            newdata = app.model.HistoryDatasetAssociation(
                create_dataset=True, sa_session=app.model.context
            )  # This import should become a library
            newdata.set_size()
            newdata.extension = file_type
            newdata.name = (
                basic_name
                + " ("
                + microbe_info[kingdom][org]["chrs"][chr]["data"][description]["feature"]
                + " for "
                + microbe_info[kingdom][org]["name"]
                + ":"
                + chr
                + ")"
            )
            app.model.context.add(newdata)
            app.model.context.flush()
            app.security_agent.copy_dataset_permissions(base_dataset.dataset, newdata.dataset)
            history.add_dataset(newdata)
            app.model.context.add(history)
            app.model.context.flush()
            try:
                copyfile(filepath, newdata.get_file_name())
                newdata.info = newdata.name
                newdata.state = newdata.states.OK
            except Exception:
                newdata.info = "The requested file is missing from the system."
                newdata.state = newdata.states.ERROR
            newdata.dbkey = dbkey
            newdata.init_meta()
            newdata.set_peek()
            app.model.context.flush()



"""Galaxy job runners to use Amazon AWS native compute resources, such as AWS Batch."""

import bisect
import hashlib
import json
import logging
import os
import re
import time
from queue import Empty
from typing import (
    Set,
    TYPE_CHECKING,
)

from galaxy import model
from galaxy.job_execution.output_collect import default_exit_code_file
from galaxy.jobs.runners import (
    AsynchronousJobRunner,
    AsynchronousJobState,
    JobState,
)
from galaxy.util import (
    smart_str,
    unicodify,
)

if TYPE_CHECKING:
    from galaxy.jobs import MinimalJobWrapper

BOTO3_IMPORT_MSG = (
    "The Python 'boto3' package is required to use "
    "this feature, please install it or correct the "
    "following error:\nImportError {msg!s}"
)

try:
    import boto3
except ImportError as e:
    boto3 = None  # type: ignore[assignment]
    BOTO3_IMPORT_MSG.format(msg=unicodify(e))


__all__ = ("AWSBatchJobRunner",)
log = logging.getLogger(__name__)

STOP_SIGNAL = object()


class AWSBatchRunnerException(Exception):
    pass


def to_dict(segments, v):
    if len(segments) == 0:
        return v
    return {segments[0]: to_dict(segments[1:], v)}


def _write_logfile(logfile, msg):
    with open(logfile, "w") as fil:
        fil.write(msg)


def _add_galaxy_environment_variables(vcpu, memory):
    # Set:
    # GALAXY_SLOTS, round 0.25 vpuc to 1.
    # GALAXY_MEMORY_MB
    return [
        {"name": "GALAXY_SLOTS", "value": str(int(max(vcpu, 1)))},
        {"name": "GALAXY_MEMORY_MB", "value": str(memory)},
    ]


def _add_resource_requirements(destination_params):
    rval = [
        {"type": "VCPU", "value": str(destination_params.get("vcpu"))},
        {"type": "MEMORY", "value": str(destination_params.get("memory"))},
    ]
    if n_gpu := destination_params.get("gpu"):
        rval.append({"type": "GPU", "value": str(n_gpu)})
    return rval


class AWSBatchJobRunner(AsynchronousJobRunner):
    """
    This runner uses container only. It requires that an AWS EFS is mounted as a local drive
    and all Galaxy job-related paths, such as objects, job_directory, tool_directory and so
    on, are placed in the EFS drive. Via this runner, Galaxy jobs are dispatched to AWS Batch
    for compute using the docker image specified by a Galaxy tool. As AWS EFS is designed to
    be able to mount at multiple places with read and write capabilities, Galaxy and Batch
    containers share the same EFS drive as a local device. Sample configurations can be found
    in `config/job_conf.sample.yml`.
    """

    runner_name = "AWSBatchRunner"
    RUNNER_PARAM_SPEC_KEY = "runner_param_specs"
    JOB_NAME_PREFIX = "galaxy-"
    # AWS Batch queries up to 100 jobs at once.
    MAX_JOBS_PER_QUERY = 100
    # Higher minimum interval as jobs are queried in batches.
    MIN_QUERY_INTERVAL = 10

    # fmt: off
    RUNNER_PARAM_SPEC = {
        "aws_access_key_id": {
            "map": str,
        },
        "aws_secret_access_key": {
            "map": str,
        },
        "region": {
            "map": str,
        },
    }

    # fmt: off
    DESTINATION_PARAMS_SPEC = {
        "vcpu": {
            "default": 1.0,
            "map": (lambda x: int(float(x)) if int(float(x)) == float(x) else float(x)),
        },
        "memory": {
            "default": 2048,
            "map": int,
        },
        "gpu": {
            "default": 0,
            "map": int,
        },
        "job_queue": {
            "default": "",
            "map": str,
            "required": True,
        },
        "job_role_arn": {
            "default": "",
            "map": str,
            "required": True,
        },
        "efs_filesystem_id": {
            "default": "",
            "map": str,
            "required": True,
        },
        "efs_mount_point": {
            "default": "",
            "map": str,
            "required": True,
        },
        "execute_role_arn": {
            "default": "",
            "map": str,
        },
        "fargate_version": {
            "default": "",
            "map": str,
        },
        "auto_platform": {
            "default": False,
            "map": lambda x: str(x).lower() == "true",
        },
        "ec2_host_volumes": {
            "default": "",
            "map": str,
        },
        "privileged": {
            "default": False,
            "map": lambda x: str(x).lower() == "true",
        },
        "retry_attempts": {
            "default": 1,
            "map": int,
        },
        "retry_on_exit_statusReason": {
            "default": "",
            "map": str,
        },
        "retry_on_exit_reason": {
            "default": "",
            "map": str,
        },
        "retry_on_exit_exitCode": {
            "default": "",
            "map": str,
        },
        "retry_on_exit_action": {
            "default": "RETRY",
            "map": str,
        },
    }

    # fmt: off
    FARGATE_RESOURCES = {
        0.25: [512, 1024, 2048],
        0.5: [1024, 2048, 3072, 4096],
        1: [2048, 3072, 4096, 5120, 6144, 7168, 8192],
        2: [4096, 5120, 6144, 7168, 8192, 9216, 10240, 11264, 12288, 13312, 14336, 15360, 16384],
        4: [8192, 9216, 10240, 11264, 12288, 13312, 14336, 15360, 16384, 17408, 18432,
            19456, 20480, 21504, 22528, 23552, 24576, 25600, 26624, 27648, 28672, 29696, 30720]
    }

    def __init__(self, app, nworkers, **kwargs):
        """Initialize this job runner and start the monitor thread"""
        assert boto3, BOTO3_IMPORT_MSG
        if self.RUNNER_PARAM_SPEC_KEY not in kwargs:
            kwargs[self.RUNNER_PARAM_SPEC_KEY] = {}
        kwargs[self.RUNNER_PARAM_SPEC_KEY].update(self.RUNNER_PARAM_SPEC)
        super().__init__(app, nworkers, **kwargs)
        session = boto3.Session(
            aws_access_key_id=self.runner_params.get("aws_access_key_id") or None,
            aws_secret_access_key=self.runner_params.get("aws_secret_access_key") or None,
            region_name=self.runner_params.get("region") or None,
        )
        self._batch_client = session.client("batch")

    def queue_job(self, job_wrapper):
        log.debug(f"Starting queue_job for job {job_wrapper.get_id_tag()}")
        if not self.prepare_job(job_wrapper, include_metadata=False, modify_command_for_container=False):
            log.debug(f"Not ready {job_wrapper.get_id_tag()}")
            return

        job_destination = job_wrapper.job_destination
        destination_params = self.parse_destination_params(job_destination.params)
        job_def = self._get_job_definition(job_wrapper, destination_params)
        job_name, job_id = self._submit_job(job_def, job_wrapper, destination_params)
        job_wrapper.set_external_id(job_id)
        ajs = AsynchronousJobState(
            files_dir=job_wrapper.working_directory,
            job_wrapper=job_wrapper,
            job_name=job_name,
            job_id=job_id,
            job_destination=job_destination,
        )
        self.monitor_queue.put(ajs)

    def _get_job_definition(self, job_wrapper, destination_params):
        tool_id = job_wrapper.tool.id
        container_image = self._find_container(job_wrapper).container_id
        h = hashlib.new("sha256")
        h.update(smart_str(container_image))
        s = json.dumps(destination_params, sort_keys=True)
        h.update(smart_str(s))
        queue_name = destination_params.get("job_queue").rsplit("/", 1)[-1]

        jd_name = f"galaxy_tool__{tool_id}__{h.hexdigest()}__{queue_name}"
        jd_name = re.sub(r'[^-_A-Za-z0-9]', '-', jd_name)[:128]
        res = self._batch_client.describe_job_definitions(jobDefinitionName=jd_name, status="ACTIVE")
        if not res["jobDefinitions"]:
            jd_arn = self._register_job_definition(jd_name, container_image, destination_params)
        else:
            jd_arn = res["jobDefinitions"][0]["jobDefinitionArn"]
            log.debug(f"Found existing job definition: {jd_name}.")

        return jd_arn

    def _get_mount_volumes(self, destination_params):
        volumes, mount_points = [], []
        volumes.append(
            {
                "name": "efs_whole",
                "efsVolumeConfiguration": {
                    "fileSystemId": destination_params.get("efs_filesystem_id"),
                    "rootDirectory": "/",
                    "transitEncryption": "ENABLED",
                    "authorizationConfig": {"iam": "ENABLED"},
                },
            },
        )
        mount_points.append(
            {
                "containerPath": destination_params.get("efs_mount_point"),
                "readOnly": False,
                "sourceVolume": "efs_whole",
            },
        )
        if destination_params.get("platform") == 'Fargate':   # Fargate doesn't support host volumes
            return volumes, mount_points

        if (ec2_host_volumes := destination_params.get("ec2_host_volumes")):
            for ix, vol in enumerate(ec2_host_volumes.split(",")):
                vol = vol.strip()
                vol_name = "host_vol_" + str(ix)
                volumes.append(
                    {
                        "name": vol_name,
                        "host": {"sourcePath": vol},
                    },
                )
                mount_points.append(
                    {
                        "containerPath": vol,
                        "readOnly": False,
                        "sourceVolume": vol_name,
                    },
                )
        return volumes, mount_points

    def _get_retry_strategy(self, destination_params):
        """ Make a simple one-condition retry strategy
        """
        # TODO make multi-condition retry strategies
        attemps = destination_params.get("retry_attempts")
        status_reason = destination_params.get("retry_on_exit_statusReason")
        reason = destination_params.get("retry_on_exit_reason")
        exit_code = destination_params.get("retry_on_exit_exitCode")
        action = destination_params.get("retry_on_exit_action")

        if attemps <= 1:
            return

        strategy = {
            "attempts": attemps,
            "evaluateOnExit": [{"action": action}],
        }
        if status_reason:
            strategy["evaluateOnExit"][0]["onStatusReason"] = status_reason
        if reason:
            strategy["evaluateOnExit"][0]["onReason"] = reason
        if exit_code:
            strategy["evaluateOnExit"][0]["onExitCode"] = exit_code
        return strategy

    def _register_job_definition(self, jd_name, container_image, destination_params):
        log.debug(f"Registering a new job definition: {jd_name}.")
        platform = destination_params.get("platform")
        volumes, mount_points = self._get_mount_volumes(destination_params)

        # TODO: support multi-node
        containerProperties = {
            "image": container_image,
            "command": [
                "/bin/sh",
            ],
            "jobRoleArn": destination_params.get("job_role_arn"),
            "executionRoleArn": destination_params.get("execute_role_arn") or destination_params.get("job_role_arn"),
            "volumes": volumes,
            "mountPoints": mount_points,
            "resourceRequirements": _add_resource_requirements(destination_params),
            "environment": _add_galaxy_environment_variables(
                destination_params.get("vcpu"), destination_params.get("memory"),
            ),
            "user": f"{os.getuid()}:{os.getgid()}",
            "privileged": destination_params.get("privileged"),
            "logConfiguration": {"logDriver": "awslogs"},
        }
        if platform == "FARGATE":
            containerProperties.update(
                {
                    "networkConfiguration": {"assignPublicIp": "ENABLED"},
                    "fargatePlatformConfiguration": {"platformVersion": destination_params.get("fargate_version")},
                    "logConfiguration": {"logDriver": "awslogs"},
                }
            )
        other_kwargs = {}
        if (retry_strategy := self._get_retry_strategy(destination_params)):
            other_kwargs["retryStrategy"] = retry_strategy

        res = self._batch_client.register_job_definition(
            jobDefinitionName=jd_name,
            type="container",
            platformCapabilities=[platform],
            containerProperties=containerProperties,
            **other_kwargs,
        )

        assert res["ResponseMetadata"]["HTTPStatusCode"] == 200
        return res["jobDefinitionArn"]

    def _submit_job(self, job_def, job_wrapper, destination_params):
        job_name = self.JOB_NAME_PREFIX + job_wrapper.get_id_tag()
        command_script_path = self.write_command(job_wrapper)

        log.info(f"Submitting job {job_name} to AWS Batch.")
        res = self._batch_client.submit_job(
            jobName=job_name,
            jobQueue=destination_params.get("job_queue"),
            jobDefinition=job_def,
            containerOverrides={
                "command": [
                    "/bin/bash",
                    f"{command_script_path}",
                ]
            },
        )

        assert res["ResponseMetadata"]["HTTPStatusCode"] == 200
        return job_name, res["jobId"]

    def stop_job(self, job_wrapper):
        job = job_wrapper.get_job()
        external_id = job.get_job_runner_external_id()
        job_name = self.JOB_NAME_PREFIX + job_wrapper.get_id_tag()

        self._batch_client.terminate_job(jobId=external_id, reason="Terminated by Galaxy!")
        msg = "Job {name!r} is terminated"
        log.debug(msg.format(name=job_name))

    def recover(self, job, job_wrapper):
        msg = "(name!r/runner!r) is still in {state!s} state, adding to the runner monitor queue"
        job_id = job.get_job_runner_external_id()
        job_name = self.JOB_NAME_PREFIX + job_wrapper.get_id_tag()
        ajs = AsynchronousJobState(
            files_dir=job_wrapper.working_directory,
            job_wrapper=job_wrapper,
            job_id=str(job_id),
            job_name=job_name,
            job_destination=job_wrapper.job_destination,
        )
        if job.state in (model.Job.states.RUNNING, model.Job.states.STOPPED):
            log.debug(msg.format(name=job.id, runner=job.job_runner_name, state=job.state))
            ajs.old_state = model.Job.states.RUNNING
            ajs.running = True
            self.monitor_queue.put(ajs)
        elif job.state == model.Job.states.QUEUED:
            log.debug(msg.format(name=job.id, runner=job.job_runner_name, state="queued"))
            ajs.old_state = model.Job.states.QUEUED
            ajs.running = False
            self.monitor_queue.put(ajs)

    def fail_job(self, job_state: JobState, exception=False, message="Job failed", full_status=None):
        job = job_state.job_wrapper.get_job()
        if getattr(job_state, "stop_job", True) and job.state != model.Job.states.NEW:
            self.stop_job(job_state.job_wrapper)
        job_state.job_wrapper.reclaim_ownership()
        self._handle_runner_state("failure", job_state)
        if not job_state.runner_state_handled:
            full_status = full_status or {}
            tool_stdout = full_status.get("stdout")
            tool_stderr = full_status.get("stderr")
            fail_message = getattr(job_state, "fail_message", message)
            job_state.job_wrapper.fail(
                fail_message, tool_stdout=tool_stdout, tool_stderr=tool_stderr, exception=exception
            )
            self._finish_or_resubmit_job(job_state, "", fail_message)
            if job_state.job_wrapper.cleanup_job == "always":
                job_state.cleanup()

    def monitor(self):
        """
        Watches jobs currently in the monitor queue and deals with state
        changes (queued to running) and job completion.
        """
        while True:
            # Take any new watched jobs and put them on the monitor list
            try:
                while True:
                    async_job_state = self.monitor_queue.get_nowait()
                    if async_job_state is STOP_SIGNAL:
                        # TODO: This is where any cleanup would occur
                        self.handle_stop()
                        return
                    self.watched.append((async_job_state.job_id, async_job_state))
            except Empty:
                pass
            # Iterate over the list of watched jobs and check state
            try:
                self.check_watched_items()
            except Exception:
                log.exception("Unhandled exception checking active jobs")
            # Sleep a bit before the next state check
            time.sleep(max(self.app.config.job_runner_monitor_sleep, self.MIN_QUERY_INTERVAL))

    def check_watched_items(self):
        done: Set[str] = set()
        self.check_watched_items_by_batch(0, len(self.watched), done)
        self.watched = [x for x in self.watched if x[0] not in done]

    def check_watched_items_by_batch(self, start: int, end: int, done: Set[str]):
        jobs = self.watched[start : start + self.MAX_JOBS_PER_QUERY]
        if not jobs:
            return

        jobs_dict = dict(jobs)
        resp = self._batch_client.describe_jobs(jobs=list(jobs_dict.keys()))

        gotten = set()
        for job in resp["jobs"]:
            status = job["status"]
            job_id = job["jobId"]
            gotten.add(job_id)
            job_state = jobs_dict[job_id]

            if status == "SUCCEEDED":
                self._mark_as_successful(job_state)
                done.add(job_id)
            elif status == "FAILED":
                reason = job["statusReason"]
                self._mark_as_failed(job_state, reason)
                done.add(job_id)
            elif status in ("STARTING", "RUNNING"):
                self._mark_as_active(job_state)
            # remain queued for "SUBMITTED", "PENDING" and "RUNNABLE"
            # TODO else?

        for job_id in jobs_dict:
            if job_id in gotten:
                continue
            job_state = jobs_dict[job_id]
            reason = f"The track of Job {job_state} was lost for unknown reason!"
            self._mark_as_failed(job_state, reason)
            done.add(job_id)

        self.check_watched_items_by_batch(start + self.MAX_JOBS_PER_QUERY, end, done)

    def _mark_as_successful(self, job_state):
        _write_logfile(job_state.output_file, "")
        _write_logfile(job_state.error_file, "")
        job_state.running = False
        self.mark_as_finished(job_state)

    def _mark_as_active(self, job_state):
        job_state.running = True
        job_state.job_wrapper.change_state(model.Job.states.RUNNING)

    def _mark_as_failed(self, job_state, reason):
        _write_logfile(job_state.error_file, reason)
        job_state.running = False
        job_state.stop_job = False
        job_state.job_wrapper.change_state(model.Job.states.ERROR)
        job_state.fail_message = reason
        self.mark_as_failed(job_state)

    def parse_destination_params(self, params):
        if not params.get("docker_enabled"):
            raise AWSBatchRunnerException("AWSBatchJobRunner needs 'docker_enabled' to be set as True!")

        check_required = []
        parsed_params = {}
        for k, spec in self.DESTINATION_PARAMS_SPEC.items():
            value = params.get(k, spec.get("default"))
            if spec.get("required") and not value:
                check_required.append(k)
            mapper = spec.get("map")
            parsed_params[k] = mapper(value)  # type: ignore[operator]
        if check_required:
            raise AWSBatchRunnerException(
                "AWSBatchJobRunner requires the following params to be provided: {}.".format(", ".join(check_required))
            )

        # parse Platform
        platform = "EC2"
        auto_platform = parsed_params.get("auto_platform")
        fargate_version = parsed_params.get("fargate_version")
        vcpu = parsed_params.get("vcpu")
        memory = parsed_params.get("memory")
        gpu = parsed_params.get("gpu")

        if auto_platform and not fargate_version:
            raise AWSBatchRunnerException("AWSBatchJobRunner needs 'farget_version' to be set to enable auto platform!")

        if gpu and (fargate_version or auto_platform):
            raise AWSBatchRunnerException(
                "GPU mode is not allowed when 'fargate_version' and/or 'auto_platform' are set!"
            )

        if fargate_version and not auto_platform:
            platform = "FARGATE"
        if auto_platform:
            fargate_vcpus = list(self.FARGATE_RESOURCES.keys())
            max_vcpu = fargate_vcpus[-1]
            max_memory = self.FARGATE_RESOURCES[max_vcpu][-1]
            if vcpu <= max_vcpu and memory <= max_memory:  # type: ignore[operator]
                c_ix = bisect.bisect_left(fargate_vcpus, vcpu)  # type: ignore[type-var]
                length = len(fargate_vcpus)
                while c_ix < length:
                    c = fargate_vcpus[c_ix]
                    m_ix = bisect.bisect_left(self.FARGATE_RESOURCES[c], memory)  # type: ignore[type-var]
                    if m_ix < len(self.FARGATE_RESOURCES[c]):
                        platform = "FARGATE"
                        parsed_params["vcpu"] = c
                        parsed_params["memory"] = self.FARGATE_RESOURCES[c][m_ix]
                        break
                    c_ix += 1
            # parse JOB QUEUE
            job_queues = parsed_params.get("job_queue").split(",")  # type: ignore[union-attr]
            if len(job_queues) < 2:
                raise AWSBatchRunnerException(
                    "AWSBatchJobRunner needs to set TWO job queues ('Farget Queue, EC2 Qeueue')"
                    " when 'auto_platform' is enabled!"
                )
            parsed_params["job_queue"] = job_queues[platform == "EC2"].strip()

        if platform == "Fargate" and parsed_params.get("ec2_host_volumes") and not auto_platform:
            raise AWSBatchRunnerException(
                "AWSBatchJobRunner: param 'ec2_host_volumes' only works for EC2 platform"
                " unless 'auto_platform' is enabled!"
            )

        parsed_params["platform"] = platform
        return parsed_params

    def write_command(self, job_wrapper: "MinimalJobWrapper") -> str:
        # Create command script instead passing it in the container
        # preventing wrong characters parsing.
        command_line = job_wrapper.runner_command_line
        job_id = job_wrapper.get_id_tag()
        job_file = JobState.default_job_file(job_wrapper.working_directory, job_id)
        exit_code_path = default_exit_code_file(job_wrapper.working_directory, job_id)
        job_script_props = {
            "command": command_line,
            "exit_code_path": exit_code_path,
            "working_directory": job_wrapper.working_directory,
            "shell": job_wrapper.shell,
            "galaxy_virtual_env": None,
        }
        job_file_contents = self.get_job_file(job_wrapper, **job_script_props)
        self.write_executable_script(job_file, job_file_contents, job_io=job_wrapper.job_io)
        return job_file



#!/usr/bin/env python
# Guruprasad Ananda
# MAQ mapper for SOLiD colourspace-reads
from __future__ import print_function

import os
import subprocess
import sys
import tempfile


def stop_err(msg):
    sys.stderr.write("%s\n" % msg)
    sys.exit()


def __main__():
    out_fname = sys.argv[1].strip()
    out_f2 = open(sys.argv[2].strip(), "r+")
    ref_fname = sys.argv[3].strip()
    f3_read_fname = sys.argv[4].strip()
    f3_qual_fname = sys.argv[5].strip()
    paired = sys.argv[6]
    if paired == "yes":
        r3_read_fname = sys.argv[7].strip()
        r3_qual_fname = sys.argv[8].strip()
    min_mapqual = int(sys.argv[9].strip())
    max_mismatch = int(sys.argv[10].strip())
    out_f3name = sys.argv[11].strip()
    subprocess_dict = {}

    ref_csfa = tempfile.NamedTemporaryFile()
    ref_bfa = tempfile.NamedTemporaryFile()
    ref_csbfa = tempfile.NamedTemporaryFile()
    cmd2_1 = "maq fasta2csfa %s > %s 2>&1" % (ref_fname, ref_csfa.name)
    cmd2_2 = "maq fasta2bfa %s %s 2>&1" % (ref_csfa.name, ref_csbfa.name)
    cmd2_3 = "maq fasta2bfa %s %s 2>&1" % (ref_fname, ref_bfa.name)
    try:
        os.system(cmd2_1)
        os.system(cmd2_2)
        os.system(cmd2_3)
    except Exception as erf:
        stop_err(str(erf) + "Error processing reference sequence")

    if paired == "yes":  # paired end reads
        tmpf = tempfile.NamedTemporaryFile()  # forward reads
        tmpr = tempfile.NamedTemporaryFile()  # reverse reads
        tmps = tempfile.NamedTemporaryFile()  # single reads
        tmpffastq = tempfile.NamedTemporaryFile()
        tmprfastq = tempfile.NamedTemporaryFile()
        tmpsfastq = tempfile.NamedTemporaryFile()

        cmd1 = "solid2fastq_modified.pl 'yes' %s %s %s %s %s %s %s 2>&1" % (
            tmpf.name,
            tmpr.name,
            tmps.name,
            f3_read_fname,
            f3_qual_fname,
            r3_read_fname,
            r3_qual_fname,
        )
        try:
            os.system(cmd1)
            os.system("gunzip -c %s >> %s" % (tmpf.name, tmpffastq.name))
            os.system("gunzip -c %s >> %s" % (tmpr.name, tmprfastq.name))
            os.system("gunzip -c %s >> %s" % (tmps.name, tmpsfastq.name))

        except Exception as eq:
            stop_err("Error converting data to fastq format." + str(eq))

        # Make a temp directory where the split fastq files will be stored
        try:
            split_dir = tempfile.mkdtemp()
            split_file_prefix_f = tempfile.mktemp(dir=split_dir)
            split_file_prefix_r = tempfile.mktemp(dir=split_dir)
            splitcmd_f = "split -a 2 -l %d %s %s" % (
                32000000,
                tmpffastq.name,
                split_file_prefix_f,
            )  # 32M lines correspond to 8M reads
            splitcmd_r = "split -a 2 -l %d %s %s" % (
                32000000,
                tmprfastq.name,
                split_file_prefix_r,
            )  # 32M lines correspond to 8M reads

            os.system(splitcmd_f)
            os.system(splitcmd_r)
            os.chdir(split_dir)
            ii = 0
            for fastq in os.listdir(split_dir):
                if not fastq.startswith(split_file_prefix_f.split("/")[-1]):
                    continue
                fastq_r = (
                    split_file_prefix_r + fastq.split(split_file_prefix_f.split("/")[-1])[1]
                )  # find the reverse strand fastq corresponding to forward strand fastq
                tmpbfq_f = tempfile.NamedTemporaryFile()
                tmpbfq_r = tempfile.NamedTemporaryFile()
                cmd3 = (
                    "maq fastq2bfq %s %s 2>&1; maq fastq2bfq %s %s 2>&1; maq map -c %s.csmap %s %s %s 1>/dev/null 2>&1; maq mapview %s.csmap > %s.txt"
                    % (
                        fastq,
                        tmpbfq_f.name,
                        fastq_r,
                        tmpbfq_r.name,
                        fastq,
                        ref_csbfa.name,
                        tmpbfq_f.name,
                        tmpbfq_r.name,
                        fastq,
                        fastq,
                    )
                )
                subprocess_dict["sp" + str(ii + 1)] = subprocess.Popen([cmd3], shell=True, stdout=subprocess.PIPE)
                ii += 1
            while True:
                all_done = True
                for j in range(len(subprocess_dict)):
                    if subprocess_dict["sp" + str(j + 1)].wait() != 0:
                        err = subprocess_dict["sp" + str(j + 1)].communicate()[1]
                        if err is not None:
                            stop_err("Mapping error: %s" % err)
                        all_done = False
                if all_done:
                    break
            cmdout = "for map in *.txt; do cat $map >> %s; done" % (out_fname)
            os.system(cmdout)

            tmpcsmap = tempfile.NamedTemporaryFile()
            cmd_cat_csmap = "for csmap in *.csmap; do cat $csmap >> %s; done" % (tmpcsmap.name)
            os.system(cmd_cat_csmap)

            tmppileup = tempfile.NamedTemporaryFile()
            cmdpileup = "maq pileup -m %s -q %s %s %s > %s" % (
                max_mismatch,
                min_mapqual,
                ref_bfa.name,
                tmpcsmap.name,
                tmppileup.name,
            )
            os.system(cmdpileup)
            tmppileup.seek(0)
            print("#chr\tposition\tref_nt\tcoverage\tSNP_count\tA_count\tT_count\tG_count\tC_count", file=out_f2)
            for line in open(tmppileup.name):
                elems = line.strip().split()
                ref_nt = elems[2].capitalize()
                read_nt = elems[4]
                coverage = int(elems[3])
                a, t, g, c = 0, 0, 0, 0
                ref_nt_count = 0
                for ch in read_nt:
                    ch = ch.capitalize()
                    if ch not in ["A", "T", "G", "C", ",", "."]:
                        continue
                    if ch in [",", "."]:
                        ch = ref_nt
                        ref_nt_count += 1
                    try:
                        nt_ind = ["A", "T", "G", "C"].index(ch)
                        if nt_ind == 0:
                            a += 1
                        elif nt_ind == 1:
                            t += 1
                        elif nt_ind == 2:
                            g += 1
                        else:
                            c += 1
                    except ValueError as we:
                        print(we, file=sys.stderr)
                print(
                    "%s\t%s\t%s\t%s\t%s\t%s" % ("\t".join(elems[:4]), coverage - ref_nt_count, a, t, g, c), file=out_f2
                )
        except Exception as er2:
            stop_err("Encountered error while mapping: %s" % (str(er2)))

    else:  # single end reads
        tmpf = tempfile.NamedTemporaryFile()
        tmpfastq = tempfile.NamedTemporaryFile()
        cmd1 = "solid2fastq_modified.pl 'no' %s %s %s %s %s %s %s 2>&1" % (
            tmpf.name,
            None,
            None,
            f3_read_fname,
            f3_qual_fname,
            None,
            None,
        )
        try:
            os.system(cmd1)
            os.system("gunzip -c %s >> %s" % (tmpf.name, tmpfastq.name))
            tmpf.close()
        except Exception:
            stop_err("Error converting data to fastq format.")

        # Make a temp directory where the split fastq files will be stored
        try:
            split_dir = tempfile.mkdtemp()
            split_file_prefix = tempfile.mktemp(dir=split_dir)
            splitcmd = "split -a 2 -l %d %s %s" % (
                32000000,
                tmpfastq.name,
                split_file_prefix,
            )  # 32M lines correspond to 8M reads
            os.system(splitcmd)
            os.chdir(split_dir)
            for i, fastq in enumerate(os.listdir(split_dir)):
                tmpbfq = tempfile.NamedTemporaryFile()
                cmd3 = (
                    "maq fastq2bfq %s %s 2>&1; maq map -c %s.csmap %s %s  1>/dev/null 2>&1; maq mapview %s.csmap > %s.txt"
                    % (fastq, tmpbfq.name, fastq, ref_csbfa.name, tmpbfq.name, fastq, fastq)
                )
                subprocess_dict["sp" + str(i + 1)] = subprocess.Popen([cmd3], shell=True, stdout=subprocess.PIPE)

            while True:
                all_done = True
                for j in range(len(subprocess_dict)):
                    if subprocess_dict["sp" + str(j + 1)].wait() != 0:
                        err = subprocess_dict["sp" + str(j + 1)].communicate()[1]
                        if err is not None:
                            stop_err("Mapping error: %s" % err)
                        all_done = False
                if all_done:
                    break

            cmdout = "for map in *.txt; do cat $map >> %s; done" % (out_fname)
            os.system(cmdout)

            tmpcsmap = tempfile.NamedTemporaryFile()
            cmd_cat_csmap = "for csmap in *.csmap; do cat $csmap >> %s; done" % (tmpcsmap.name)
            os.system(cmd_cat_csmap)

            tmppileup = tempfile.NamedTemporaryFile()
            cmdpileup = "maq pileup -m %s -q %s %s %s > %s" % (
                max_mismatch,
                min_mapqual,
                ref_bfa.name,
                tmpcsmap.name,
                tmppileup.name,
            )
            os.system(cmdpileup)
            tmppileup.seek(0)
            print("#chr\tposition\tref_nt\tcoverage\tSNP_count\tA_count\tT_count\tG_count\tC_count", file=out_f2)
            for line in open(tmppileup.name):
                elems = line.strip().split()
                ref_nt = elems[2].capitalize()
                read_nt = elems[4]
                coverage = int(elems[3])
                a, t, g, c = 0, 0, 0, 0
                ref_nt_count = 0
                for ch in read_nt:
                    ch = ch.capitalize()
                    if ch not in ["A", "T", "G", "C", ",", "."]:
                        continue
                    if ch in [",", "."]:
                        ch = ref_nt
                        ref_nt_count += 1
                    try:
                        nt_ind = ["A", "T", "G", "C"].index(ch)
                        if nt_ind == 0:
                            a += 1
                        elif nt_ind == 1:
                            t += 1
                        elif nt_ind == 2:
                            g += 1
                        else:
                            c += 1
                    except Exception:
                        pass
                print(
                    "%s\t%s\t%s\t%s\t%s\t%s" % ("\t".join(elems[:4]), coverage - ref_nt_count, a, t, g, c), file=out_f2
                )
        except Exception as er2:
            stop_err("Encountered error while mapping: %s" % (str(er2)))

    # Build custom track from pileup
    chr_list = []
    out_f2.seek(0)
    fcov = tempfile.NamedTemporaryFile()
    fout_a = tempfile.NamedTemporaryFile()
    fout_t = tempfile.NamedTemporaryFile()
    fout_g = tempfile.NamedTemporaryFile()
    fout_c = tempfile.NamedTemporaryFile()
    fcov.write(
        """track type=wiggle_0 name="Coverage track" description="Coverage track (from Galaxy)" color=0,0,0 visibility=2\n"""
    )
    fout_a.write(
        """track type=wiggle_0 name="Track A" description="Track A (from Galaxy)" color=255,0,0 visibility=2\n"""
    )
    fout_t.write(
        """track type=wiggle_0 name="Track T" description="Track T (from Galaxy)" color=0,255,0 visibility=2\n"""
    )
    fout_g.write(
        """track type=wiggle_0 name="Track G" description="Track G (from Galaxy)" color=0,0,255 visibility=2\n"""
    )
    fout_c.write(
        """track type=wiggle_0 name="Track C" description="Track C (from Galaxy)" color=255,0,255 visibility=2\n"""
    )

    for line in out_f2:
        if line.startswith("#"):
            continue
        elems = line.split()
        chr = elems[0]

        if chr not in chr_list:
            chr_list.append(chr)
            if not (chr.startswith("chr") or chr.startswith("scaffold")):
                chr = "chr"
            header = "variableStep chrom=%s" % (chr)
            fcov.write("%s\n" % (header))
            fout_a.write("%s\n" % (header))
            fout_t.write("%s\n" % (header))
            fout_g.write("%s\n" % (header))
            fout_c.write("%s\n" % (header))
        try:
            pos = int(elems[1])
            cov = int(elems[3])
            a = int(elems[5])
            t = int(elems[6])
            g = int(elems[7])
            c = int(elems[8])
        except Exception:
            continue
        fcov.write("%s\t%s\n" % (pos, cov))
        try:
            a_freq = a * 100.0 / cov
            t_freq = t * 100.0 / cov
            g_freq = g * 100.0 / cov
            c_freq = c * 100.0 / cov
        except ZeroDivisionError:
            a_freq = t_freq = g_freq = c_freq = 0
        fout_a.write("%s\t%s\n" % (pos, a_freq))
        fout_t.write("%s\t%s\n" % (pos, t_freq))
        fout_g.write("%s\t%s\n" % (pos, g_freq))
        fout_c.write("%s\t%s\n" % (pos, c_freq))

    fcov.seek(0)
    fout_a.seek(0)
    fout_g.seek(0)
    fout_t.seek(0)
    fout_c.seek(0)
    os.system(
        "cat %s %s %s %s %s | cat > %s" % (fcov.name, fout_a.name, fout_t.name, fout_g.name, fout_c.name, out_f3name)
    )


if __name__ == "__main__":
    __main__()
